/*
 * Copyright (C) 1998-2002 Wolfgang Moser <womo_(at)_d81_(dot)_de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Commodore CBM 1581 floppy disk copy util for PC's, 1581HLVL.C
 */

#include "1581hlvl.h"
#include <stdio.h>
#include <conio.h>
#include <time.h>
#include <assert.h>

const char *joberr[5]={"Write ","Verify ","Read ","Format ","Format'n'write "};

static int CBMPreRWV(uint8 drive, uint8 cyl, uint8 side, uint tries,
						   uint8 *secID, uint8 *sideID);

/*
	  80,	// Cylinders:   80 cylinders per disk
	   2,	// Heads:        2 sides (tracks) per cylinder
	  10,	// Sectors:     10 sectors per track (side)
	   2,	// BpS:        512 bytes per sector
	   2,	// Rate:       250 kilobit per second
	   1,	// Modulation: MFM
	0xDF,	// SpecifyCMD, step rate time: 0x0D, head load time: 0x0F
	0x0C,	// RWGAP3:      12
	0x23,	// FmtGAP:      35
	0x00	// Pattern:      0
*/
static const FDCparameterSet FDprmSets[FDUNDEF]={
	{00,0,00,0,0,0,0x00,0x00,0x00,0x00},	// FDUNKNW
	{80,2,10,2,2,1,0xDF,0x0C,0x23,0x00},	// CBM1581
   {81,2,10,3,0,1,0xCF,0x0C,0x64,0x00}		// CMDFD2M
//	{81,2,10,3,0,1,0xCF,0x0C,0x64,0xE5}		// GAP and Pattern are from the CMD FD2000 manual
	};


// waits for a disk or a new disk in the specified drive
//
// input:  drive:		the drive, where a disk has to be inserted into
//         newDisk:  if not 0, a new disk is required (disk change)
// output: 1 - disk or new disk is in
//         0 - the wait has been aborted by the user (<ESC> pressed)
//
int waitForDiskChng(uint8 drive, int newDisk){
	int key=0x20,o,n;

	if(newDisk) printf("Please change the disk ...\n");
   else printf("Waiting for disk drive ready status, please insert a disk ...\n");

   o=-1;
	do{
   	if(key==0x00) while(kbhit()) getch();	// eat up special combos
	   do{
      	n=senseDiskStatus(drive);
         if(o!=n){
         	printf("\rDisk Status: ");
				switch(n){
	         	case 0: printf("disk is out     "); break;
   	         case 1: printf("(old) disk is in"); break;
      	      case 2: printf("new disk is in  ");
	         	}
            }
		   switch(n){
            case 1:						// (old) disk is in
         	   if(newDisk) break;	// if new disk insert needed, don't return
      	   case 2:						// new disk
					printf("\n");
         	   return 1;
      	   }
#ifdef TimeSlice      	   
			releaseTmSlice();
#endif
         o=n;
   	   }while(!kbhit());
      }while((key=getch())!=0x1B);
	printf("\n");
	return 0;
	}

// prints all values from a result string to the console
//
// input:  pointer to the result string structure
// output: none
//
void CBMprintStatus(resultString *resStr, const char *Job, int Retrying){
   if(resStr==NULL) return;

   if(resStr->D.ST1&0x02) Job="Write protect ";
   else if(Job==NULL) Job="";

	if(resStr->D.ST0&0xFF00){
   	if(resStr->D.ST0>=-2) printf("FDC doesn't answer (%sput timeout).\n",(resStr->D.ST0)==-1?"in":"out");
      else printf("Unknown FDC communication problem\n");
      }
	else printf("0x%02X,0x%02X,0x%02X,%s%s,%2d,%1d,%2d,%1d%s\n",
					resStr->D.ST0, resStr->D.ST1, resStr->D.ST2,
               Job,	resStr->D.ST0&0xC0?"ERROR":"OK",
					resStr->D.Cylinder+1, resStr->D.Side,
					resStr->D.Sector, resStr->D.BytesPS,
					Retrying&&(resStr->D.ST0&0xC0)?"    retrying ...":"");
	}

// reset and recalibrate the specified drive
//
// input:  drive:		the drive, that shall be resetted
// output: SUCCESS - the operation was successful
//         GENFAIL - the operation was unsucessful
//
ENo CBMresetDrive(uint8 drive){
	resetDrive(drive);
	return (recalibrateDrive(drive,0)==NULL)?SUCCESS:GENFAIL;
	}

// recalibrate the specified drive
//
// input:  drive:		the drive, that shall be recalibrated
// output: SUCCESS - the operation was successful
//         GENFAIL - the operation was unsucessful
//
ENo CBMrecalibrate(uint8 drive){
	resultString *RES;
   signed char i;


   for(i=2;i>=0;i--){
		RES=recalibrateDrive(drive,0);
		if(RES==NULL) break;
		CBMprintStatus(RES,"Recalibrate ",i);
      }


	if(RES!=NULL && CBMresetDrive(drive)==GENFAIL){
		printf("Unsuccessful attempt to recalibrate drive %d\n",drive);
//		CBMprintStatus(RES);
		return GENFAIL;
		}
	return SUCCESS;
	}

// perform a read/write head seek of the specified drive
//
// input:  drive:		the drive, where the seek shall be done
//         track:		the cylinder number, where the seek shall be done
// output: SUCCESS - the operation was successful
//         GENFAIL - the operation was unsucessful
//
ENo CBMseek(uint8 drive, uint8 cyl){
	resultString *RES; //,RS;
	uint8 t=cyl-1;
	signed char retry=3;

	while((RES=seekToCyl(drive,t,0))!=NULL){
		CBMprintStatus(RES,"Seek ",retry);
		if(retry--<=0 || CBMrecalibrate(drive)==GENFAIL){
			printf("Unsuccessful attempt to seek to track %2d on drive %d\n",cyl,drive);
//			CBMprintStatus(RES);
			return GENFAIL;
			}
		}
	return SUCCESS;
	}

// sets the floppy disk parameter table to the selected type
//
// input:  drive:		the drive, that is beeing presetted
//         Set:		the name (enum constant) of the predefined parameter set
//         StepRate:	the step rate time, parametrized with 1/4 ms (with the range
//                   from 0.25ms (0x01) ... 63,75ms (0xFF), 0x00 - don't change)
// input:  Set:	the name (enum constant) of the predefined parameter set
// output: none
//
void CBMpresetFDprm(uint8 drive, FDprmSet Set, uint8 StepRate){
	uint16 Rate;

	if(Set<CBM1581 || Set>=FDUNDEF) return;		// unknown type, don't change anything

	assert(FDprmSets[Set].Sectors<=MaxSectors);
		// increase sector buffer parameter via CENTRAL.H, if assertion fault
	assert((128u << FDprmSets[Set].BpS)*FDprmSets[Set].Sectors<=TrBufSize);
		// increase track buffer size via CENTRAL.H, if assertion fault
	FDprm=FDprmSets[Set];
	if(StepRate) do{
		Rate=StepRate-1;		// 8-bit/16-bit conversion
			// the real value is dependend from the selected bitrate
		switch(FDprm.Bitrate){
			default:
				continue;	// big break to the end of do{}while;
			case 1:		//  300 kBit-Selection
				Rate=(Rate*3+2)/20;
				break;
			case 3:		// 1000 kBit-Selection
				Rate>>=1;
			case 0:		//  500 kBit-Selection
				Rate>>=1;
			case 2:		//  250 kBit-Selection
				Rate>>=1;
			}
		Rate=0x0F-Rate;
		if(Rate>0x0F) Rate=0x00;
		FDprm.SpecifyCMD=(Rate<<4) | (FDprm.SpecifyCMD&0x0F);	// set rate

		}while(0);
	CBMrecalibrate(drive);
	}


// checks, if the inserted disk is a CBM formatted disk and detects the format
//
// input:  drive:		the drive, where the disk to be checked is in
//         chkSwap:  0 - don't check for swapped side IDs
//                   other - check, if sectors on head 1 (side 0)
//                           contain sector IDs that are 0
// output: NON_3_5 -	the drive type is no 3,5" type
//         CBM1581 - the disk was detected as CBM 1581 disk format compatible
//         CMDFD2M - the disk was detected as CMD FD2000 disk format compatible
//         FDUNDEF - the disk was not detected as a CBM format type
//
FDprmSet	CBMcheckDisk(uint8 drive, int chkSwap){
	resultString RS;
	FDCparameterSet oldFDprm=FDprm;	// save current parameter set

   FDprmSet	PSet;
	char		secMap[MaxSectors];		// MaxSectors is defined in CENTRAL.H

   uint8 	wrongID,type,doBR=0xFF;
	int   	i,tmo;

   	// do a CMOS check, if it is a 3,5" drive
      // if not, return -1 (can never be a "correct" disk)
   _disable();
	outp8(0x70, 0x10);	// CMOS address 0x10
	type=inp8(0x71);		// read addressed CMOS value
   _enable();
   switch(drive){
   	case 0:	type>>=4;	break;	// test upper nibble
      case 1:	type&=0x0F;	break;	// test lower nibble
      default:	type=0x04;				// assume 1,44 MB drive
   	}

   if(type<3 || type>5) return NON_3_5;	// can't support these disk types (5,25")
   if(type==3) doBR&=~1;		// 500 kbps are not supported on 720 kB disk drives

   for(PSet=FDUNDEF-1;PSet>FDUNKNW;PSet--){
   	if(doBR&(1<<FDprmSets[PSet].Bitrate)){	// if the wished bitrate is supported
			CBMpresetFDprm(drive,PSet,0);			// Set bitrate, modulation, specifyCMD
															// and standard step rate time
			if(CBMrecalibrate(drive)==SUCCESS){	// track 0 should be selected
            for(i=0;i<FDprm.Sectors;i++) secMap[i]=0;	// delete sector detection map
            wrongID=0;


               // wait for more than 4 timer ticks (222 ms, a disk round)
   	      i=5;
			   tmo=peek8(0x40, 0x6C);
            while(readSectorID(drive,0,&RS,1)){
				   if(tmo!=peek8(0x40, 0x6C)){
					   tmo=peek8(0x40,0x6C);
                  if(i--<0) break;
                  }
           	   if(RS.D.Cylinder!=0) wrongID|=1;	// TrackHeader wrong
           	   if(chkSwap && RS.D.Side!=0)     wrongID|=2;	// SideHeader wrong
				   if(RS.D.Sector<1 || RS.D.Sector>FDprm.Sectors)
						                  wrongID|=4;	// SectorHeader wrong
				   else secMap[RS.D.Sector-1]=1;		// mark sector read
           	   if(RS.D.BytesPS!=FDprm.BpS)
											   wrongID|=8;	// BpSHeader wrong
               }
            if(i>=5){	// readSectorID timeout happened
					   // deselect the desired bitrate
#if (MESSAGES >= 5)
					fprintf(stderr,"Wrong bitrate, deselecting %d\n",FDprmSets[PSet].Bitrate);
#endif
			      doBR&=~(1<<FDprmSets[PSet].Bitrate);
					   // reset controller without recalibrating again
				   resetDrive(drive);
               }
			   else{
		         for(i=0;i<FDprm.Sectors;i++) if(!secMap[i]) break;
   		      if(!wrongID && i>=FDprm.Sectors) break;	// format is detected
#if (MESSAGES >= 5)
					fprintf(stderr,"Format is not detected: 0x%02X, %d of %d\n",wrongID,i,FDprm.Sectors);
#endif
				   }
   			}
			}
      }
	FDprm=oldFDprm;		// restore parameter set (should be drive related)
	CBMrecalibrate(drive);
   return PSet;
	}

// return the sector number of the next available sector header
//
// input:  drive:		the drive, where the disk to read is in
//         cyl:		the cylinder number of the track to read
//         side:		the disk side number of the track to read
//         tries:		the number of retries for this operation
//         secID:    a pointer to the sector ID result
//         sideID:   a pointer to the side ID result
//
// output: SUCCESS - the operation was successful
//         GENFAIL - the operation failed
//         CBMFRMT - CBM incompatible disk detected
//         USRABRT - the operation was aborted by user interaction
//         NO_DISK - FDC didn't answer, no disk in drive
//
int CBMPreRWV(uint8 drive, uint8 cyl, uint8 side, uint tries,
				  uint8 *secID, uint8 *sideID){
	resultString RS, *RES;
	signed char retr1, retr2, retr3, retr4, retr5, op;

   if(secID==NULL || sideID==NULL) return GENFAIL;
	op=-1;
	retr1=retr2=retr3=retr4=retr5=tries;


	while(retr1>0 && retr2>0 && retr3>0 && retr4>0 &&retr5>0){
   	if(kbhit()){
      	switch(getch()){
            case 0x1B: return USRABRT;
         	case 0x00: while(kbhit()) getch();	// eat up special combos
         	}
      	}

		do{
			if(op--==4){
				resetDrive(drive);
				selectDrive(drive);
				if((RES=seekToCyl(drive,1,0))!=NULL){
					CBMprintStatus(RES,"Seek ",retr1);
					retr1--;
					break;
					}
				}

			op=1;
			if((RES=seekToCyl(drive,cyl,0))!=NULL){
				CBMprintStatus(RES,"Seek ",retr2);
				retr2--;
				break;
				}

			op=2;
         RES=&RS;
			if(!readSectorID(drive,side,&RS,OP_TIMEOUT)){
				CBMprintStatus(&RS,"Read header ID ",retr3);
				if(RS.D.ST0&0xFF00 && senseDiskStatus(drive)!=1) return NO_DISK;
				retr3--;
				break;
				}
#if (MESSAGES >= 5)
			fprintf(stderr,"Drive %d, Track %2d, Side %2d   --> ReadID: T %2d, H %d, S %2d, BPS %d\n",
					  drive, cyl, side, RS.D.Cylinder, RS.D.Side, RS.D.Sector, RS.D.BytesPS);
#endif
			op=3;
			if(RS.D.Cylinder!=cyl){
				retr4--;
				break;
				}
			*secID=RS.D.Sector;
  	      *sideID=RS.D.Side;
			if(/*RS.D.Side!=side || */RS.D.Sector<1 || RS.D.Sector>FDprm.Sectors || RS.D.BytesPS!=FDprm.BpS){
				printf("Possibly no CBM compatible disk inserted\n");
            return CBMFRMT;
				}
			return SUCCESS;

			}while(0);

     	if(RES!=NULL && RES->D.ST0<0) break; // timeout!!!

		if((RES=recalibrateDrive(drive,0))!=NULL){
			CBMprintStatus(RES,"Recalibrate ",retr5);
			op=4;
			retr5--;
			}
		}
	if(RES!=NULL && RES->D.ST0<0) printf("FDC is not responding on ");
	else printf("%d unsuccessful attempts to ",tries);
	switch(op){
		case 1: printf("seek to");					break;
		case 2: printf("read sector ID from");	break;
		case 3: printf("get a valid ID for");       break;
		case 4: printf("recalibrate");
		}
	if(op<4) printf(" side %1d of cylinder %2d of",side,cyl);
	printf(" drive %d\n",drive);
	return GENFAIL;
	}

// transfer a track with the multiple sector FDC feature
//
// input:  xfer:		the transfer command (Read, Write, Write&Verify)
//         drive:		the drive, where the disk to is in
//         cyl:		the cylinder number of the track to handle
//         side:		the disk side number of the track to handle
//         buffer:   a pointer to the data transfer buffer
//         ileave:	an interleave factor to be used
//         retries:	the number of retries for this operation
// output: SUCCESS - the operation was successful
//         GENFAIL - the operation failed
//         USRABRT - the operation was aborted by user interaction
//         WRPRTCT - write protect error
//         NO_DISK - FDC didn't answer, no disk in drive
//
// (Note: CBMxferTrackSS will not be invoked automatically anymore)
//
ENo CBMxferTrackSS(xferM xfer, uint8 drive, uint8 cyl, uint8 side,
						 uint8 huge *buffer, uint8 ileave, uint retries){
	const char  job[3]={'w','v','r'};
	// const char *joberr[3]={"Write ","Verify ","Read "};
	resultString *RES;
	uint8 huge *buf;
	uint retry;
	uint8 v,r,i,t=cyl-1,csec,oldcs,sideID=side;
	signed char Done[MaxSectors];		// MaxSectors is defined in CENTRAL.H
   	// markers: 0 		- to write
		//          1 		- to verify
		//          2 		- to read
		//          3 		- read/write/verify error
		//			  -1		- done
	do{
	   switch(xfer&(WrVrfy|Read)){
	      case 0:
	      	r=0;
	      	if(xfer&Format) continue;
	      default:
	      	break;
	   	case Read:
	      	if(xfer&Format) break;
	      	r=2;
	         continue;
	      case Write:
	      case WrVrfy:
	      	r=0;
	         continue;
	      case Verify:
	      	r=1;
	         continue;
	   	}
	   printf("CBMxferTrackSS: Womo isn idiot, please report this!!!\n");
		return USRABRT;
		}while(0);
	for(csec=0;csec<FDprm.Sectors;csec++) Done[csec]=r;

  		// clear done buffer
	ileave%=FDprm.Sectors;

  	if(xfer&Format){
  		if(CBMseek(drive,cyl)==GENFAIL) return GENFAIL;
  		csec=0;
      // sideID is preinitialised with side
  		}
  	else{
	      // sideID is preinitialised with side for the case of an abort
		switch(CBMPreRWV(drive,t,side,retries,&csec,&sideID)){
      	case USRABRT: return USRABRT;
         case NO_DISK: return NO_DISK;
			/*
         case CBMFRMT:
         case GENFAIL:
			*/
      	}
  		}

	retry=retries;

   while(retry-->0){
		if(kbhit()){
   		switch(getch()){
      		case 0x1B: return USRABRT;
         	case 0x00: while(kbhit()) getch();	// eat up special combos
	         }
	   	}

		if(xfer&Format){
		   printf("%2d %1d f  \r",cyl,side);
#if (MESSAGES >= 2)
   		fprintf(stderr,"%2d %1d f      \n",cyl,side);
#endif
			RES=formatCBMtrack(drive,t,side);
         if(RES!=NULL){
	         if(RES->D.ST1&0x02){
					CBMprintStatus(RES,"Format ",0);
	         	return WRPRTCT;
	         	}
				CBMprintStatus(RES,"Format ",retry);
				if(RES->D.ST0&0xFF00 && senseDiskStatus(drive)!=1) return NO_DISK;
				if(CBMrecalibrate(drive)==GENFAIL || CBMseek(drive,cyl)==GENFAIL) break;
				continue;
   	   	}

         switch(xfer&(WrVrfy|Read)){
   	      case Read:
      	      r=2;
               break;
            case Write:
            case WrVrfy:
      	      r=0;
               break;
            case Verify:
      	      r=1;
   	      }
	      for(csec=0;csec<FDprm.Sectors;csec++) Done[csec]=r;

			if(xfer&WrVrfy) csec=FDprm.Sectors-1;
			else return SUCCESS;
         }

      v=(xfer&Verify)?1:0;						// ending with verify?
      r=(xfer&Write)?0:1;						// starting with writing?
      if(xfer&Read) v=r=2;						// or is it reading
      for(;r<=v;r++){							// r=0 - write, r=1 - verify
		   printf("%2d %1d %c  \r",cyl,side,job[r]);

			for(i=0;i<FDprm.Sectors;i++){		// one round verify or write
		   	if(kbhit()){
   		   	switch(getch()){
      		      case 0x1B: return USRABRT;
         			case 0x00: while(kbhit()) getch();	// eat up special combos
	         		}
	   	   	}
            oldcs=csec;	// store the old csec for the case of not having
            				// to do more sectors in the current loop

	         	// calculate the following sector number
            csec+=ileave;
            csec%=FDprm.Sectors;
            	// search the next sector to do (dependend on write or verify)
               // if reading or writing, we search for a todo entry (0)
               // if verifying after writing, we search for a written entry (1)
				// searching the rest of the array, if nothing can be found,
				// the old csec has to be restored!
            for(;csec<FDprm.Sectors;csec++) if(Done[csec]==r) break;
            if(csec>=FDprm.Sectors){				// nothing found, search from the beginning
					// searching from the beginning of the array
					for(csec=0;csec<FDprm.Sectors;csec++) if(Done[csec]==r) break;
	            if(csec>=FDprm.Sectors){
		            csec=oldcs;
						break;	// nothing found, nothing more to do
                  }
            	}
					// next sector to do is: csec
               // calculating buf pointer
				buf=buffer+csec*(128u << FDprm.BpS);
#if (MESSAGES >= 2)
				fprintf(stderr,"%2d %1d %c %3d  \n",cyl,side,job[r],csec+1);
#endif
            switch(r){
            	case 0:	// writing
						RES=writeCBMsector(drive,t,side,csec+1,sideID,buf);
               	break;
            	case 1:	//	verifying
						RES=verifyCBMsector(drive,t,side,csec+1,sideID,buf);
               	break;
            	case 2:	// reading
						RES=readCBMsector(drive,t,side,csec+1,sideID,buf);
               	break;
               default:
               	RES=NULL;
            	}
            if(RES!=NULL){
		         if(RES->D.ST1&0x02){
						CBMprintStatus(RES,joberr[r],0);
		         	return WRPRTCT;
		         	}
					CBMprintStatus(RES,joberr[r],retry);
					if(RES->D.ST0&0xFF00 && senseDiskStatus(drive)!=1) return NO_DISK;
					Done[csec]=3;
               }
            else{	// success
            	if(r<v) Done[csec]=1;	// writing done, verify follows
            	else Done[csec]=-1;		// Done
            	}
   	   	}
      	}
   		// set all not done jobs (not -1) in the done buffer to to do (0)
      	// if all sectors are done, we can return

   	// markers: 0      - to do
		//          0x10   - read/write/verify error
		//          WrVrfy - written, if verify is done after
		//          other  - done

      r=1;
     	if(v==1 && (xfer&Write)) v=0;	// writing follows on verify, set to "do write"
      for(i=0;i<FDprm.Sectors;i++) if(Done[i]!=-1){
      	Done[i]=v;	// if there are not done or errornous sectors,
  	      r=0;			// we do a retry
      	}
      if(r) return SUCCESS;
   	}
	printf("%d unsuccessful attempts to transfer data at side %1d, cylinder %2d, drive %d\n",retries,side,cyl,drive);
   return GENFAIL;
	}

// transfer a track with the multiple sector FDC feature
//
// input:  xfer:		the transfer command (Read, Write, Write&Verify)
//         drive:		the drive, where the disk to is in
//         cyl:		the cylinder number of the track to handle
//         side:		the disk side number of the track to handle
//         buffer:   a pointer to the data transfer buffer
//         ileave:	an interleave factor to be used
//         retries:	the number of retries for this operation
// output: SUCCESS - the operation was successful
//         GENFAIL - the operation failed
//         USRABRT - the operation was aborted by user interaction
//         WRPRTCT - write protect error
//         NO_DISK - FDC didn't answer, no disk in drive
//
// (Note: CBMxferTrackSS will not be invoked automatically anymore)
//
ENo CBMxferTrackMS(xferM xfer, uint8 drive, uint8 cyl, uint8 side,
						 uint8 huge *buffer, uint8 ileave, uint retries){
	resultString *RES;
   // const char *joberr[4]={"Read ","Write ","Verify ","Format "};

	uint retry;
	uint8 t=cyl-1,csec,sideID=side;
	int job;

  	if(xfer&0x01 && xfer&0x0E){
	   printf("CBMxferTrackMS: Womo isn idiot, please report this!!!\n");
		return USRABRT;
      }

		ileave+=FDprm.Sectors-1;	// interleave is decremented by 1 for multiple
		ileave%=FDprm.Sectors;		// sector operations (default 0)
   	  									// This _can_ increase combined fromatting speed

	if(xfer&Format){
	  	if(CBMseek(drive,cyl)==GENFAIL) return GENFAIL;
      }
   else{
		switch(CBMPreRWV(drive,t,side,retries,&csec,&sideID)){
      	case USRABRT: return USRABRT;
         case NO_DISK: return NO_DISK;
/*
         case CBMFRMT:
         case GENFAIL:
*/
      	}

		csec+=ileave;
      csec++;
		while(csec>FDprm.Sectors) csec-=FDprm.Sectors;
   	}

	retry=retries;

   while(retry-->0){
	   if(kbhit()){
   	   switch(getch()){
      	   case 0x1B: return USRABRT;
         	case 0x00: while(kbhit()) getch();	// eat up special combos
	         }
   	   }
	   printf("%2d %1d ",cyl,side);
		if(xfer&Read){
		   printf("r  \r");
#if (MESSAGES >= 2)
	   	fprintf(stderr,"%2d %1d r %3d  \n",cyl,side,csec);
#endif
			job=2;
      	RES=readCBMtrack(drive,t,side,csec,sideID,buffer);
      	}
      else{
         RES=NULL;
			if(xfer&Format){
			   printf("f  \r");
#if (MESSAGES >= 2)
	   		fprintf(stderr,"%2d %1d f      \n",cyl,side);
#endif
				job=4;
				RES=formatCBMtrack(drive,t,side);
				if(xfer&WrVrfy){
               csec=1+ileave;
				   printf("%2d %1d ",cyl,side);
            	}
				}
			if(RES==NULL && xfer&Write){
			   printf("w  \r");
#if (MESSAGES >= 2)
	   		fprintf(stderr,"%2d %1d w %3d  \n",cyl,side,csec);
#endif
				job=0;
				RES=writeCBMtrack(drive,t,side,csec,sideID,buffer);
            }
			if(RES==NULL && xfer&Verify){
				if(xfer&Write){
					csec+=ileave;
					if(csec>FDprm.Sectors) csec-=FDprm.Sectors;
				   printf("%2d %1d ",cyl,side);
            	}
			   printf("v  \r");
#if (MESSAGES >= 2)
	   		fprintf(stderr,"%2d %1d v %3d  \n",cyl,side,csec);
#endif
				job=1;
				RES=verifyCBMtrack(drive,t,side,csec,sideID,buffer);
         	}
      	}
		if(RES==NULL) return SUCCESS;
      else{
         if(RES->D.ST1&0x02){
				CBMprintStatus(RES,joberr[job],0);
         	return WRPRTCT;
         	}
			CBMprintStatus(RES,joberr[job],retry);
			if(RES->D.ST0&0xFF00 && senseDiskStatus(drive)!=1) return NO_DISK;
         if(job==4 && (CBMrecalibrate(drive)==GENFAIL || CBMseek(drive,cyl)==GENFAIL)) break;
         }

		csec+=ileave;
		if(csec>FDprm.Sectors) csec-=FDprm.Sectors;
   	}
	printf("%d unsuccessful attempts to transfer data at side %1d, cylinder %2d, drive %d\n",retries,side,cyl,drive);
//	CBMprintStatus(RES);
	return GENFAIL;
	}

// format'n'write a track via intel's extended command
//
// input:  xfer:		the transfer command (Read, Write, Write&Verify)
//         drive:		the drive, where the disk to is in
//         cyl:		the cylinder number of the track to handle
//         side:		the disk side number of the track to handle
//         buffer:   a pointer to the data transfer buffer
//         retries:	the number of retries for this operation
// output: SUCCESS - the operation was successful
//         GENFAIL - the operation failed
//         USRABRT - the operation was aborted by user interaction
//         WRPRTCT - write protect error
//         NO_DISK - FDC didn't answer, no disk in drive
//
ENo CBMfwvTrack(xferM xfer, uint8 drive, uint8 cyl, uint8 side,
						 uint8 huge *buffer, uint retries){
	resultString *RES;
   //   const char *joberr[3]={"Format'n'write ","Verify "};

	uint retry;
	uint8 t=cyl-1;
	int job;

   	// operations allowed:
      // format'n'write
      // format'n'write with verify
	if((xfer&~Verify)!=(Format|Write)){
	   printf("CBMfwvTrack: Womo isn idiot, please report this!!!\n");
		return USRABRT;
      }

	if(CBMseek(drive,cyl)==GENFAIL) return GENFAIL;

	retry=retries;

   while(retry-->0){
	   if(kbhit()){
   	   switch(getch()){
      	   case 0x1B: return USRABRT;
         	case 0x00: while(kbhit()) getch();	// eat up special combos
	         }
   	   }
	   printf("%2d %1d fw \r",cyl,side);

      RES=NULL;
#if (MESSAGES >= 2)
	   fprintf(stderr,"%2d %1d fw     \n",cyl,side);
#endif
		job=4;
		RES=fnwriteCBMtrack(drive,t,side,buffer);

		if(RES==NULL && xfer&Verify){
			printf("%2d %1d v  \r",cyl,side);
#if (MESSAGES >= 2)
	   	fprintf(stderr,"%2d %1d v      \n",cyl,side);
#endif
			job=1;
			RES=verifyCBMtrack(drive,t,side,1,side,buffer);
      	}
		if(RES==NULL) return SUCCESS;
      else{
         if(RES->D.ST1&0x02){
				CBMprintStatus(RES,joberr[job],0);
         	return WRPRTCT;
         	}
			CBMprintStatus(RES,joberr[job],retry);
			if(RES->D.ST0&0xFF00 && senseDiskStatus(drive)!=1) return NO_DISK;
         if(job==4 && (CBMrecalibrate(drive)==GENFAIL || CBMseek(drive,cyl)==GENFAIL)) break;
         }
   	}
	printf("%d unsuccessful attempts to format'n'write at side %1d, cylinder %2d, drive %d\n",retries,side,cyl,drive);
	return GENFAIL;
	}

// detects the existence of an external 48 MHz oscillator for intel
// 82078-64pin controllers and fixes the FDC to operate at 24MHz
//
// input:  drive:		the drive, where a disk has to be inserted into
// output: 0 - 48MHz oscillator was detected but couldn't be fixed
//         1 - external 48MHz oscillator was not detected
//         2 - external 48MHz oscillator was detected and fixed
//
int handle82078_48(uint8 drive){
	int i;
   clock_t op24;

	CBMpresetFDprm(drive, CBM1581, 0);	// set the default

	i=0;
	for(;;){
		op24=clock();
		while(clock()==op24);	// wait for the next change
		CBMseek(drive, 10);
		CBMrecalibrate(drive);
		op24=clock()-op24;
		op24--;						// compensate the change waiting

		if(op24>1)	break;		// compensate external 48 MHz oscillator
		
		if(i++){
			printf("Fixing the 82078-64pin external 48 MHz oscillator failed.\n");
			return 0;				// switching failed
			}
		switch48MHz(1);			// and check the timing again
		}
	return i+1;
	}
