/*
 * Copyright (C) 1998-2002 Wolfgang Moser <womo_(at)_d81_(dot)_de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Commodore CBM 1581 floppy disk copy util for PC's, 1581HLVL.H
 */

#ifndef __CBM_1581_High_Level__
#define __CBM_1581_High_Level__

#include "fdclowlv.h"

typedef enum{
	Read  = 0x01,
	Write = 0x02,
	Verify= 0x04,	// Verify only is currently not used
	WrVrfy= 0x06,
	Format= 0x08
	}xferM;

// these are name constants for the different array indexes
typedef enum{
	NON_3_5=-1,		// the disk drive is no 3,5" one
	FDUNKNW= 0,		// unknown floppy disk format

	CBM1581= 1,		// CBM1581 compatible floppy disk
   CMDFD2M= 2,		// CMD FD2000 compatible floppy disk

   FDUNDEF= 3		// must be the last entry (undefined)
	}FDprmSet;
/*
	  80,	// Cylinders:   80 cylinders per disk
	   2,	// Heads:        2 sides (tracks) per cylinder
	  10,	// Sectors:     10 sectors per track (side)
	   2,	// BpS:        512 bytes per sector
	   2,	// Rate:       250 kilobit per second
	   1,	// Modulation: MFM
	0xDF,	// SpecifyCMD
	0x0C,	// RWGAP3:      12
	0x23,	// FmtGAP:      35
	0x00	// Pattern:      0
*/

// general error numbers for the high level routines
typedef enum{
   NO_DISK=-4,		// FDC didn't answer, because there's no disk
   WRPRTCT=-3,		// write protect error
	USRABRT=-2,		// the user aborted the operation
	CBMFRMT=-1,		// CBM incompatible disk detected
	GENFAIL= 0,		// the operation was not successful
	SUCCESS= 1		// operation completed successful
	}ENo;

extern const FDCparameterSet FDprmSets[FDUNDEF];

int		waitForDiskChng(uint8 drive, int newDisk);

void		CBMprintStatus	(resultString *resStr, const char *Job, int Retrying);

ENo		CBMresetDrive	(uint8 drive);
ENo		CBMrecalibrate	(uint8 drive);
ENo		CBMseek			(uint8 drive, uint8 cyl);

void		CBMpresetFDprm	(uint8 drive, FDprmSet Set, uint8 StepRate);
FDprmSet	CBMcheckDisk	(uint8 drive, int chkSwap);

ENo		CBMxferTrackSS	(xferM xfer, uint8 drive, uint8 cyl, uint8 side,
								 uint8 huge *buffer, uint8 ileave, uint retries);
ENo		CBMxferTrackMS	(xferM xfer, uint8 drive, uint8 cyl, uint8 side,
								 uint8 huge *buffer, uint8 ileave, uint retries);

ENo		CBMfwvTrack		(xferM xfer, uint8 drive, uint8 cyl, uint8 side,
						 		 uint8 huge *buffer, uint retries);
int		handle82078_48	(uint8 drive);
#endif
