/*
 * Copyright (C) 1998-2002 Wolfgang Moser <womo_(at)_d81_(dot)_de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Commodore CBM 1581 floppy disk copy util for PC's, OPERATNS.C
 */

#include "operatns.h"
#if (MESSAGES >= 3)
#include <string.h>
#endif

static void						checkBuffer		(int CacheSize);
static ENo						readOffset		(uint32 offset, uint16 size,
														 unsigned char **BufAdr);
void								markCBM_Block	(unsigned long int baseNoffset);
#ifdef FD2000BAMSupport
static unsigned long int	cv24toul			(unsigned char vs[3]);
static uint8					getMaxSector	(uint8 track);
static int						processCMD_BAM	(unsigned long int base);
static int						process1541BAM	(unsigned long int base);
static int						process1571BAM	(unsigned long int base);
#endif
static int						process1581BAM	(unsigned long int base);

unsigned char *TransferBuffer=NULL;					// was: buffer
	// 3 times the maximum track size to avoid DMA 64kB page overflows
static unsigned char		GenBuf[3*TrBufSize];
static unsigned char	  *TBufButterflyPtr=NULL;

FILE			  *FileDescriptor=NULL;					// was: fd

unsigned int	FloppyDriveNo=0;						// was: drive
unsigned int	MultipleDisks=0;						// was: Multiple
unsigned int	SingleSectorC=0;						// was: MSecRead
	// the following three would be better combined into one "xferM" variable
FDprmSet			FormattingDsk=FDUNKNW;				// was: format
xferM				WritingToDisk=0;						// was: writing
xferM				VerifyWriting=0;						// was: verify

unsigned int	FrmtNWrt82078=0;						// was: FnW
unsigned int	NumberRetries=3;						// was: tries
unsigned int	SectorIntrlve=1;						// was: ileave

unsigned char *BAMcopymarker;							// BAM copy track markers
																// (max. 168 tracks)
static char	  *FilenameRetStr;						// output path and filename

// initialise the track cache buffer, ensure,
// that it doesn't cross a 64kB page boundary
//
// input:  none
// output: none
//         
void initBufferPointers(void){
	unsigned long distance;
	unsigned char *BPtr;

#if (TrBufSize==0) || (3ul*TrBufSize>0x10000ul)
#error Buffer setup: programming error "TrBufSize" is out of range
#endif

		// calculate the next following 64kB page boundary
	distance=((unsigned long)FP_SEG(GenBuf)<<4)+FP_OFF(GenBuf);
		// and calculate the distance to the start of the buffer then
	distance=(~distance+1)&0xFFFF;

	TBufButterflyPtr=GenBuf;
	switch(distance/TrBufSize){
		case 0:	// the first of 3 buffers crosses a 64kB page boundary
			TBufButterflyPtr+=2*TrBufSize;
			BPtr=GenBuf;	// The first buffer is free
			break;
		case 1:	// the second of 3 buffers crosses a 64kB page boundary
			TBufButterflyPtr+=(int)distance;
					// get the bigger one of the remaining half buffers
			if(distance>(3ul*TrBufSize)/2){
				BPtr=GenBuf;	// The first buffer half is free
				}
			else{
					// the ending of the buffer is bigger
				BPtr=GenBuf+(unsigned int)((5ul*TrBufSize)/2);
				}
			break;
		case 2:	// the third of 3 buffers crosses a 64kB page boundary
		default:	// none of the buffers crosses a 64kB page boundary
			TBufButterflyPtr+=TrBufSize;
			BPtr=TBufButterflyPtr+TrBufSize;	// The last buffer is free
		}

#if (MESSAGES >= 1)
	fprintf(stderr, "Buffer setup: case: %ld distance 0x%08lX GenBuf 0x%05lX end 0x%05lX\n"
	       "              TBufButterflyPtr 0x%05lX start 0x%05lX end 0x%05lX\n"
	       "              Remaining buffer setup: start 0x%05lX end 0x%05lX\n",
	       distance/TrBufSize, distance,
          ((unsigned long)FP_SEG(GenBuf)<<4)+FP_OFF(GenBuf),
          ((unsigned long)FP_SEG(GenBuf)<<4)+FP_OFF(GenBuf)+3ul*TrBufSize,
          ((unsigned long)FP_SEG(TBufButterflyPtr)<<4)+FP_OFF(TBufButterflyPtr),
          ((unsigned long)FP_SEG(TBufButterflyPtr)<<4)+FP_OFF(TBufButterflyPtr)-TrBufSize,
          ((unsigned long)FP_SEG(TBufButterflyPtr)<<4)+FP_OFF(TBufButterflyPtr)+TrBufSize,
          ((unsigned long)FP_SEG(BPtr)<<4)+FP_OFF(BPtr),
          ((unsigned long)FP_SEG(BPtr)<<4)+FP_OFF(BPtr)+TrBufSize/2
          );
#endif
		// The rest of the three buffers may be divided into two parts,
		// the bigger one contains at least TrBufSize/2 bytes
#if (FilepathLength+BAMcpmark>TrBufSize/2)
#error Buffer setup: rest of buffers is not big enough
#endif
	FilenameRetStr=(char *)BPtr;
	BPtr+=FilepathLength;
	BAMcopymarker=BPtr;
	}

// check the buffer and also check, if the current size is correct
//
// input:  the size of the cache buffers for the current FDprm setting
// output: none
//         
void checkBuffer(int CacheSize){
	if(TBufButterflyPtr==NULL){
		printf("\n\n::checkBuffer: transfer buffer butterfly pointer isn't initialised\n\n");
		exit1581Copy(RetTBOR,-1);
		}
	if(CacheSize<=0 || CacheSize>TrBufSize){
		printf("\n\n::checkBuffer: current image format capacity is out of range\n\n");
		exit1581Copy(RetTBOR,-1);
		}
	}

// return the adress of the indicated single track buffer 
//
// input:  No:	the position to read the bytes from
// output: the address of the buffer
//         
unsigned char *getTrackCache(int No){
	int cBSize, bufs;

		// read out FDprm and calculate the buffer size needed
		// as well as the maximum index number possible
	cBSize=(128u << FDprm.BpS)*FDprm.Sectors;
	checkBuffer(cBSize);
	bufs=TrBufSize/cBSize;
	
	if(No<0 || No>=2*bufs){
		printf("\n\n::getTrackCache: buffer index is out of range\n\n");
		exit1581Copy(RetTBOR,-1);
		}
#if (MESSAGES >= 3)
	fprintf(stderr, "::getTrackCache: index %d, buffer range: 0x%05lX - 0x%05lX\n",
	       No,
          ((unsigned long)FP_SEG(TBufButterflyPtr)<<4)+FP_OFF(TBufButterflyPtr)+cBSize*(No-bufs),
          ((unsigned long)FP_SEG(TBufButterflyPtr)<<4)+FP_OFF(TBufButterflyPtr)+cBSize*(1+No-bufs)
          );
#endif
	return TBufButterflyPtr+cBSize*(No-bufs);
	}

// abort the execution and exit 1581-Copy
//
// input:  retcode - tells the return code (as enum)
//         stdmsg  - tells the abort message
//
void exit1581Copy(ExitCode retcode, unsigned int stdmsg){
	const char *msgs[]={
	/* 0 */ "User abort.       \n",
	/* 1 */ "File reading error.\n",
	/* 2 */ "File writing error.\n",
	/* 3 */ "Filename generation error.\n",
	/* 4 */ "Filename extension generation error.\n",
	/* 5 */ "\nWrong drive type detected, this is no 3,5\" floppy, aborting.\n",
   /* 6     no message */
   /* -1    surely no message */
		};

	if(stdmsg<6) printf(msgs[stdmsg]);

	CBMrecalibrate(FloppyDriveNo);
	switchMotorOff();
	if(FileDescriptor!=NULL) fclose(FileDescriptor);
   exit(retcode);
	}

// transfer track between TransferBuffer and disk track
//
// input:  xfer:	the transfer job and direction to do
//         cyl:	the cylinder number of the track to be transferred
//         side:	the side number of the track to be transferred
// output: SUCCESS - the operation was successful
//         GENFAIL - the operation failed
//         USRABRT - the operation was aborted by user interaction
//         WRPRTCT - write protect error
//         NO_DISK - FDC didn't answer, no disk in drive
//
ENo transferDiskTrack(xferM xfer ,uint8 cyl, uint8 side){
	int res=GENFAIL;	// use normal FormattingDsk and/or write, if not Format'n'Writing or an error occurs

	if(FrmtNWrt82078 && (xfer&(Write|Format))==(Write|Format)){
#ifdef ForceFnWVerify
		res=CBMfwvTrack(xfer|Verify,FloppyDriveNo,cyl,side,TransferBuffer,NumberRetries);
#else
		res=CBMfwvTrack(xfer,FloppyDriveNo,cyl,side,TransferBuffer,NumberRetries);
#endif
		}
	if(res==GENFAIL){
		if(!SingleSectorC){
			res=CBMxferTrackMS(xfer,FloppyDriveNo,cyl,side,TransferBuffer,SectorIntrlve,NumberRetries);
			}
		if(res==GENFAIL){
			res=CBMxferTrackSS(xfer,FloppyDriveNo,cyl,side,TransferBuffer,SectorIntrlve,NumberRetries);
			}
		}
   return res;
	}

// read a block (track) from disk or file
//
// input:  offset:	the position to read the bytes from
//         size:		the minimum size of the bytes to read
// output: SUCCESS - the operation was successful
//         GENFAIL - the operation failed
//         USRABRT - the operation was aborted by user interaction
//         NO_DISK - FDC didn't answer, no disk in drive
//
ENo readOffset(uint32 offset, uint16 size, unsigned char **BufAdr){
	ENo res;
	uint8 t, maxT;

	// *BufAdr=CylinderBuffer;
	*BufAdr=getTrackCache(0);
	if(size>(128u << FDprm.BpS)*FDprm.Sectors){
		printf("readOffset: checking size, internal error\n");
		return USRABRT;
		}
	if(offset+size>((128ul << FDprm.BpS)*FDprm.Sectors*FDprm.Heads*FDprm.Cylinders)){
		printf("readOffset: checking offset, internal error\n");
		return USRABRT;
		}

#if (MESSAGES >= 3)
	printf("readOffset: file reading at least %d bytes from offset: 0x%08lX\n", size, offset);
#endif
	if(WritingToDisk){	// reading BAM from file
		fseek(FileDescriptor, offset, SEEK_SET);
		if(fread(*BufAdr, 1, (128u << FDprm.BpS)*FDprm.Sectors, FileDescriptor)<size){
			return GENFAIL;
			}
      }
	else{
			// calculate track and side number(s) to read
		maxT=(offset+size)/((128ul << FDprm.BpS)*FDprm.Sectors);
		t=offset/((128ul << FDprm.BpS)*FDprm.Sectors);	// doubled: 0...161
		TransferBuffer=*BufAdr;

		do{
#if (MESSAGES >= 3)
			printf("readOffset: reading Disktrack: T(%d), S(%d)\n", t/FDprm.Heads+1,t%FDprm.Heads);
#endif
			res=transferDiskTrack(Read,t/FDprm.Heads+1,t%FDprm.Heads);
			if(res<=GENFAIL) return res;

			t++;
			TransferBuffer+=FDprm.Sectors*(128u << FDprm.BpS);
			}while(t<maxT);

		*BufAdr+=(uint16)(offset%((128ul << FDprm.BpS)*FDprm.Sectors));
   	}
	return SUCCESS;
	}

// mark a BAM block of the current FDprmSet via offset numbers
// input:  none
// output: none
//
void markCBM_Block(unsigned long int baseNoffset){
	baseNoffset/=((128ul << FDprm.BpS)*FDprm.Sectors);		// calc track
	BAMcopymarker[(uint16)(baseNoffset>>3)]|=1<<(baseNoffset&0x07);	// mark track
	}

#ifdef FD2000BAMSupport
// get the maximum sector number for a track of 35- or 70-track CBM images
// input:  track - the track number to get the maximum sector number from
//                 (0...34/69)
// output: the number of sectors this tracks can hold
//
uint8 getMaxSector(uint8 track){
	if(track>34) track-=35;
	if(track>23){
		if(track>29)   return 17;
		else           return 18;
		}
	else if(track>16) return 19;
	else              return 21;
	}

unsigned long int cv24toul(unsigned char vs[3]){
	unsigned long int val;
	
	val=vs[0];
	val<<=8;
	val|=vs[1];
	val<<=8;
	val|=vs[2];
	return val;
	}

// read the BAM of a native CMD partition
// input:  base:	the base address of the native partition to read the BAM from
// output:  1 - the operation was successful
//          0 - the operation failed (not used currently)
//
int processCMD_BAM(unsigned long int base){
	unsigned char *bbp;
	unsigned int i;
	unsigned int maxSecs;

		// read the whole BAM into the buffer
	if(readOffset(base+0x200L, 0x2000, &bbp)!=SUCCESS) return 0;
	maxSecs=bbp[8]*256;	// max tracks with 256 sectors each

	bbp+=0x20;
#if (MESSAGES >= 3)
	printf("Native partition consecutive BAM (256 sectors at up to 255 tracks):\n");
#endif

	for(i=0;i<maxSecs;i++){
		if(!(bbp[i>>3]&(0x80>>(i&0x07)))){
				// calculate offset and call BAMmark with this...
			markCBM_Block(base+256ul*i);
#if (MESSAGES >= 3)
			printf("*");
			}
		else{
			printf("-");
#endif
			}
		}
#if (MESSAGES >= 3)
	printf("\n");
#endif
	return 1;
	}

// read the BAM of a virtual 1541 disk
// input:  base:	the base address of the virtual disk to read the BAM from
// output:  1 - the operation was successful
//          0 - the operation failed (not used currently)
//
int process1541BAM(unsigned long int base){
	uint8 t, s, maxS;
	unsigned char *bbp;
	unsigned long int offset;

		/* add BAM offset to base and load the desired size into the buffer */
	if(readOffset(base+0x16500L, 0x100, &bbp)!=SUCCESS) return 0;


		// 35 tracks with up to 21 sectors
#if (MESSAGES >= 3)
	printf(" S:           11111111112\n"
	       "T : 012345678901234567890");
#endif
	offset=0l;
	bbp++;
	for(t=0;t<35;t++){
		bbp+=4;
#if (MESSAGES >= 3)
		printf("\n%2d: ", t+1);
#endif
			// getMaxSector(Track)
		maxS=getMaxSector(t);

		for(s=0;s<maxS;s++){
			if(!(bbp[s>>3]&(1<<(s&0x07)))){
					// calculate offset and call BAMmark with this...
				markCBM_Block(base+offset);
#if (MESSAGES >= 3)
				printf("*");
				}
			else{
				printf("-");
#endif
				}
			offset+=0x100;
			}
		}
#if (MESSAGES >= 3)
	printf("\nLast offset: 0x%06lX  %ld\n", offset, offset);
#endif
	return 1;
	}

// read the BAM of a virtual 1571 disk
// input:  base:	the base address of the virtual disk to read the BAM from
// output:  1 - the operation was successful
//          0 - the operation failed (not used currently)
//
int process1571BAM(unsigned long int base){
	uint8 t, s, maxS;
	unsigned char *bbp;
	unsigned long int offset;
		// the first part of the BAM works like the 1541 BAM
	if(!process1541BAM(base)) return 0;
		/* add BAM offset to base and load the desired size into the buffer */
	if(readOffset(base+0x41000L, 0x100, &bbp)!=SUCCESS) return 0;

		// tracks 36...70 with up to 21 sectors
	offset=174848l;
	for(t=35;t<70;t++){
#if (MESSAGES >= 3)
		printf("\n%2d: ", t+1);
#endif
			// getMaxSector(Track)
		maxS=getMaxSector(t);

		for(s=0;s<maxS;s++){
			if(!(bbp[s>>3]&(1<<(s&0x07)))){
					// calculate offset and call BAMmark with this...
				markCBM_Block(base+offset);
#if (MESSAGES >= 3)
				printf("*");
				}
			else{
				printf("-");
#endif
				}
			offset+=0x100;
			}
		bbp+=3;
		}
#if (MESSAGES >= 3)
	printf("\nLast offset: 0x%06lX  %ld\n", offset, offset);
#endif
	return 1;
	}
#endif

// read the BAM of a virtual 1581 disk
// input:  base:	the base address of the virtual disk to read the BAM from
// output:  1 - the operation was successful
//          0 - the operation failed (not used currently)
//
int process1581BAM(unsigned long int base){
	uint8 t, s;
	unsigned char *bbp;

		/* add BAM offset to base and load the desired size into the buffer */
	if(readOffset(base+0x61800L, 0x400, &bbp)!=SUCCESS) return 0;

		// 80 Tracks a 40 Sektoren
#if (MESSAGES >= 3)
	printf(" S:           111111111122222222223333333333\n"
	       "T : 0123456789012345678901234567890123456789");
#endif
	bbp+=0x111-6;
	for(t=0;t<80;t++){
		if(t==40) bbp+=6+0x10;	// bufpointer korrigieren
		else bbp+=6;
#if (MESSAGES >= 3)
		printf("\n%2d: ", t+1);
#endif
		for(s=0;s<40;s++){
			if(!(bbp[s>>3]&(1<<(s&0x07)))){
					// calculate offset and call BAMmark with this...
				markCBM_Block(base+(40ul*t+s)*256);
#if (MESSAGES >= 3)
				printf("*");
				}
			else{
				printf("-");
#endif
				}
			}
		}
#if (MESSAGES >= 3)
	printf("\n");
#endif
	return 1;
	}

// read the BAM of a disk or image into the track selector
// input:  FDprmSet:	the type of the disk/image to read the BAM(s) from
// output:  1 - the operation was successful
//          0 - the operation failed (not used currently)
//         -1 - the operation was aborted by user interaction
int getTrackAllocation(FDprmSet type){
#ifdef FD2000BAMSupport
	int i, res;
	unsigned char *bp;

   unsigned int TableMax;
	typedef struct{
		unsigned char type;
		unsigned long int offset, size;
#if (MESSAGES >= 3)
		char Name[17];
#endif
		}ParEntry;
	ParEntry ParTable[32];
#else
	int i;
#endif

	do{
		for(i=0;i<BAMcpmark;i++) BAMcopymarker[i]=0x00;
	
		switch(type){
			case CBM1581:
				// call 1581 marker with offset 0
				if(!process1581BAM(0ul)) continue;
				break;
	   	case CMDFD2M:
#ifdef FD2000BAMSupport
		   		// read the system partition, 3rd sector of track 81, 1024 bytes
				res=readOffset((128ul << FDprm.BpS)*(2+80*FDprm.Sectors*FDprm.Heads), 1024, &bp);
				switch(res){
					case NO_DISK:
					case WRPRTCT:
					case CBMFRMT:
					case GENFAIL:
					default:
						continue;
					case USRABRT:
						return -1;
					case SUCCESS:
						;
					}
					// always mark system partition itself
				BAMcopymarker[160>>3]|=1<<(160&0x07);
				
					// bp holds the correct pointer now...
		   		// process parition table
				for(i=0,TableMax=0;i<32;i++,TableMax++, bp+=32){
					ParTable[i].type=bp[2];
					ParTable[i].offset=cv24toul(bp+0x15)<<9;
					ParTable[i].size=cv24toul(bp+0x1d)<<9;
#if (MESSAGES >= 3)
					strncpy(ParTable[i].Name, (char *)(bp+5), 16);
					ParTable[i].Name[16]='\0';
#endif
					}
				for(i=0;i<TableMax;i++){
#if (MESSAGES >= 3)
					printf("No %2d TP: 0x%02X Name: \"%16s\" Off: 0x%08lX Sz: %6ld Bl: %4ld\n",
						i, ParTable[i].type, ParTable[i].Name, ParTable[i].offset, ParTable[i].size, ParTable[i].size>>8);
#endif
					switch(ParTable[i].type){
						case 1:		/* native */
								// ParTable[x].size is not needed, because native
								// partitions know a "last track used" descriptor
								// Byte $08 of BAM
							res=processCMD_BAM(ParTable[i].offset);
							break;
						case 2:		/* 1541 emulation ??? */
							res=process1541BAM(ParTable[i].offset);
							break;
						case 3:		/* 1571 emulation ??? */
							res=process1571BAM(ParTable[i].offset);
							break;
						case 4:		/* 1581 emulation ??? */
							res=process1581BAM(ParTable[i].offset);
							break;
						default:		/* type error */
							res=0;
						}
					if(!res){
							// from offset up to size...
						for(res=(uint16)(ParTable[i].offset/((128ul << FDprm.BpS)*FDprm.Sectors));
								res<(ParTable[i].offset+ParTable[i].size)/((128ul << FDprm.BpS)*FDprm.Sectors);
								res++){
							BAMcopymarker[res>>3]|=1<<(res&0x07);
							}
						}
					}
				break;
#endif
			default:			// mark all tracks to be copied
				printf("BAM copying of this disk type is not supported yet.\n");
			case NON_3_5:	// mark all tracks to be copied
			case FDUNKNW:
   		case FDUNDEF:
				continue;
			}
#if (MESSAGES >= 1)
		printf("BAM copy track table is:\n 1...40/0: ");
		for(i=0;i<80;i+=2) printf("%c", (BAMcopymarker[i>>3]&(1<<(i&0x07)))?'*':'-');
		printf("\n       /1: ");
		for(i=1;i<80;i+=2) printf("%c", ((BAMcopymarker[i>>3]&(1<<(i&0x07)))?'*':'-'));
		printf("\n41...80/0: ");
		for(i=80;i<160;i+=2) printf("%c", (BAMcopymarker[i>>3]&(1<<(i&0x07)))?'*':'-');
		printf("\n       /1: ");
		for(i=81;i<160;i+=2) printf("%c", ((BAMcopymarker[i>>3]&(1<<(i&0x07)))?'*':'-'));
		printf("\n81...84/0: ");
		for(i=160;i<168;i+=2) printf("%c", (BAMcopymarker[i>>3]&(1<<(i&0x07)))?'*':'-');
		printf("\n       /1: ");
		for(i=161;i<168;i+=2) printf("%c", ((BAMcopymarker[i>>3]&(1<<(i&0x07)))?'*':'-'));
		printf("\n");
#endif
		return 1;
		}while(0);
	for(i=0;i<BAMcpmark;i++) BAMcopymarker[i]=0xFF;
	return 0;
	}

// write default system partition and/or BAM to disk
//
// output: SUCCESS - the operation was successful
//         GENFAIL - the operation failed
//         USRABRT - the operation was aborted by user interaction
//         WRPRTCT - write protect error
//         NO_DISK - FDC didn't answer, no disk in drive
//
int writeDefaultDisk(void){
	int res;

	TransferBuffer=getTrackCache(0);
   switch(FormattingDsk){
      case CBM1581:
			printf("Writing 1581 default BAM to disk in drive %c:\n", 'A'+FloppyDriveNo);

			C1581createBAM(TransferBuffer);

			res=transferDiskTrack(Write|VerifyWriting,40,0);
         break;
      case CMDFD2M:
			printf("Writing FD2000 default system partition to disk in drive %c:\n", 'A'+FloppyDriveNo);

			FD2000createSYS(TransferBuffer);

			if((res=transferDiskTrack(Write|VerifyWriting,81,0))<GENFAIL) break;
			printf("Writing FD2000 default BAM to disk in drive %c:\n", 'A'+FloppyDriveNo);
			FD2000createBAM(TransferBuffer);

			res=transferDiskTrack(Write|VerifyWriting,1,0);
         break;
      default:
	      printf("Writing a default BAM to this disk type is not supported yet.\n");
         res=1;
      }
	return res;
   }

// generate a new indexed filename from the previous one
//
// input:  inStr:		the old indexed or unindexed file name
// output: not NULL: the new indexed file name
//         NULL:		a new indexed file name couldn't be generated
//
char *buildInxName(char *inStr){
	const int maxLength=8;

	register int i;
	int length,delimiter,indexStr;
	unsigned long int inx;
	char outStr[13], inxStr[11], *ptr, *op, *rs;

	if(inStr==NULL) return NULL;

   rs=FilenameRetStr;
   // separate filename from path and copy path into retString
   	// search end of string and copy path
	for(ptr=inStr,op=FilenameRetStr;*ptr!='\0';ptr++,op++) *op=*ptr;
   	// search last ':', '\' or '/'
   while(ptr>=inStr && *ptr!=':' && *ptr!='\\' && *ptr!='/') ptr--;
   if(ptr>=inStr){	// there has been a path delimiter found
   	ptr++;			// the following char is the first of the filename

			// subtraction of two pointers results in an integer
         // the problem is, that it is a long int, but the first
         // pointer "retStr" is no far pointer, so that the cast
			// is needed (perhaps this causes a bug?)
      rs=FilenameRetStr+(int)(ptr-inStr);
      inStr=ptr;
   	}

	delimiter=-1;
	for(i=0,ptr=inStr;*ptr!='\0';i++,ptr++) if(*ptr=='.') delimiter=i;
	length=i;
	if(delimiter==-1) delimiter=length;

#if (MESSAGES >= 3)
		fprintf(stderr,"Delimiter: \"%s\"\n",inStr+delimiter);
#endif

		// get index string part (up to a maximum of 8 chars)
	for(i=delimiter,ptr=inStr+delimiter-1;ptr>=inStr;i--,ptr--)
		if(*ptr<'0' || *ptr>'9' || delimiter-i>=maxLength)
			break;
	indexStr=i;

#if (MESSAGES >= 3)
	fprintf(stderr,"Index: \"%s\"\n",inStr+indexStr);
#endif

		// build new index number string
	if(delimiter==indexStr){
			// new index number is 1
		inxStr[0]='1';
		inxStr[1]='\0';
#if (MESSAGES >= 3)
		fprintf(stderr,"no old index, generating \"1\"\n");
#endif
		}
	else{
   	inx=0L;
   	for(ptr=inStr+indexStr,i=indexStr;i<delimiter;ptr++,i++){
      	inx*=10L;
         inx+=(*ptr-'0');
      	}
//		inx=strtoul(inStr+indexStr,NULL,10);
#if (MESSAGES >= 3)
		fprintf(stderr,"old index number: \"%lu\"\n",inx);
#endif
		inx++;												// build next index number
		sprintf(inxStr, "%lu", inx);	// write index number to string
#if (MESSAGES >= 3)
		fprintf(stderr,"new index number: \"%lu\"\n",inx);
#endif
		}

		// determine the length of the new index string
	for(i=0,ptr=inxStr;*ptr!='\0';i++,ptr++);
	length=i;
#if (MESSAGES >= 3)
	fprintf(stderr,"new index length: \"%d\"\n",i);
#endif
	if(length>maxLength) return NULL;		// new filename will be invalid

		// copying base name to outStr
	for(i=0,ptr=inStr,op=outStr;i<delimiter && i<maxLength;i++,ptr++,op++) *op=*ptr;

		// new combined string will be to long
	i-=length;
	if(i<indexStr) i=indexStr;			// don't eat up file name chars, if not needed
	if(i+length>=maxLength) i=maxLength-length;
	op=outStr+i;

		// copy index number string onto the base name
	for(ptr=inxStr;i<maxLength && *ptr!='\0';i++,ptr++,op++) *op=*ptr;

		// copy extension onto file name
   for(ptr=inStr+delimiter;*ptr!='\0';ptr++,op++) *op=*ptr;

		// setting string end delimiter
	*op='\0';

   	// copy outString into static return String, so the string can
      // againg be the input of this function without conflicts
	for(ptr=outStr,op=rs,i=0;i<maxLength+5;ptr++,op++,i++) *op=*ptr;

	return FilenameRetStr;
	}

// generate a filename with the forced extension
//
// input:  inStr:		the input filename
//         force:    if true, extension changing is forced
// output: not NULL: the new generated file name
//         NULL:		a new file name couldn't be generated
//
char *changeExtension(char *inStr, int force, FDprmSet Set){
	const char *Extensions[FDUNDEF]={"","D81","D2M"};
	char *ptr, *op;

	if(inStr==NULL) return NULL;

	if(Set<CBM1581 || Set>CMDFD2M){	// permanent array bound violation check :)
#if (MESSAGES >= 3)
		fprintf(stderr,"Set (%d) invalid, extension not changed, it's \"%s\"\n",Set,inStr);
#endif
		return inStr;	// let it be unchanged
      }

	   // separate filename from path and copy path into retString
   	// search end of string and copy path
	for(ptr=inStr,op=FilenameRetStr;*ptr!='\0';ptr++,op++) *op=*ptr;
   *op='\0';

		   	// search last ':', '\' or '/' in the output string
	while(op>=FilenameRetStr && *op!=':' && *op!='\\' && *op!='/') op--;
   op++;		// select the following char of the delimiter

   ptr=NULL;// search the last delimiting dot '.', ptr marks the dot
	while(*op!='\0'){
   	if(*op=='.') ptr=op;
      op++;
   	}
   if(ptr!=NULL){		// if a dot was found don't change anything
      if(!force){		// except if we are forced to
#if (MESSAGES >= 3)
			fprintf(stderr,"Dot found, extension not changed, it's \"%s\"\n",inStr);
#endif
			return inStr;	// let it be unchanged
         }
		op=ptr;
      }
   *op++='.';
   ptr=(char *)Extensions[Set];
#if (MESSAGES >= 3)
	fprintf(stderr,"Changing extension to \"%s\"\n",ptr);
#endif

   while(*ptr!='\0') *op++=*ptr++;

#if (MESSAGES >= 3)
	fprintf(stderr,"Filename changed to \"%s\"\n",FilenameRetStr);
#endif
	return FilenameRetStr;
	}
