;;; @file c128.s
;;; Commodore 128 definitions
;;; @author Marko Mkel (marko.makela@nic.funet.fi)
host = 128

;;; start address for the server program
serverstart = $1300
;;; start address for the utility programs
utilstart = $2000

#include "c64common.s"

bstart	= $2d	; start of basic program text
bend	= $1210 ; end of basic program text

bank	= $a9	; temporary variable that holds the memory bank
stack	= $105	; offset to the stack while in interrupt

peek	= $2a2
peekptr = $2aa
poke	= $2af
pokeptr = $2b9

linkprg = $4f4f
basrun	= $5aa6

	;; exit from an interrupt
#mac irqexit
	jmp $ff33
#endm
	;; disable timer interrupts
#mac deactirq
	sei
#endm
	;; re-enable timer interrupts
#mac reactirq
#endm
	;; store the speed settings and enable fast speed
#mac phfast
	lda $d030
	pha
	lda #$1
	sta $d030	; switch to 2 MHz mode
#endm
	;; restore original speed from fast speed
#mac plfast
	pla
	sta $d030	; restore original speed
#endm
	;; store the speed settings and enable slow speed (no badlines)
#mac phslow
	lda $d011
	and #$7f	; clear bit 8 of raster counter
	pha
	lda $d015
	pha
	lda $d030
	pha
	lda #$0
	sta $d015	; disable the sprites
	sta $d011	; blank the screen
	sta $d030	; switch to 1 MHz mode
nobad$:	bit $d011	; wait for vertical blank area
	bpl nobad$
#endm
	;; restore original speed from slow speed
#mac plslow
	pla
	sta $d030	; restore the speed
	pla
	sta $d015	; restore the sprite configuration
	pla
	sta $d011	; restore the screen mode
#endm
