;;; @file c64.s
;;; Commodore 64 definitions
;;; @author Marko Mkel (marko.makela@nic.funet.fi)
host = 64

;;; start address for the server program
serverstart = $cc00
;;; start address for the utility programs
utilstart = $c000

#include "c64common.s"
bstart	= $2b	; start of basic program text
bend	= $2d	; end of basic program text
#if ramexp
bank	= $a9	; temporary variable that holds the memory bank
#else
bank	= none	; non-banked systems
#endif
blnsw	= $cc
crsw	= $d0

stack	= $104	; offset to the stack while in interrupt

clr	= $a659
linkprg = $a533
stxpt	= $a68e
newstt	= $a7ae

	;; exit from an interrupt
#mac irqexit
	jmp $ea81
#endm
	;; disable timer interrupts
#mac deactirq
	lda #$2
	sta $dc0d
#endm
	;; re-enable timer interrupts
#mac reactirq
	jsr $fddd
#endm
	;; store the speed settings and enable fast speed
#mac phfast
	lda $d011
	pha
	lda #$0
	sta $d011	; blank the screen
#endm
	;; restore original speed from fast speed
#mac plfast
	pla
	sta $d011	; restore original screen mode
#endm
	;; store the speed settings and enable slow speed (no badlines)
#mac phslow
	lda $d011
	pha
	lda $d015
	pha
	lda #$0
	sta $d015	; disable the sprites
	sta $d011	; blank the screen
nobad$:	bit $d011	; wait for vertical blank area
	bpl nobad$
#endm
	;; restore original speed from slow speed
#mac plslow
	pla
	sta $d015	; restore the sprite configuration
	pla
	sta $d011	; restore the screen mode
#endm
