/**
 * @file ihex.h
 * Intel hex input and output
 * @author Marko Mkel (msmakela@nic.funet.fi)
 */

/* Copyright  2002,2003 Marko Mkel.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

struct spi;

/** Construct hex dump from source
 * @param file	the output file
 * @param spi	the communication primitives and parameters
 * @param len	number of bytes to dump (start from zero)
 * @param src	the data source
 * @return	zero on success; nonzero on error
 */
int
ihex_output (FILE* file,
	     const struct spi* spi,
	     unsigned len,
	     int (*src) (const struct spi*,
			 unsigned, unsigned char*));

/** Write to dest from hex dump
 * @param file	the input file
 * @param spi	the communication primitives and parameters
 * @param dest	the data destination
 * @param done	function to invoke at the end of the dump (0=none)
 * @return	zero on success; nonzero on error
 */
int
ihex_input (FILE* file,
	    const struct spi* spi,
	    int (*dest) (const struct spi*,
			 unsigned, unsigned char),
	    int (*done) (const struct spi*));
