/*
 *  cload.c  v1.1
 *
 *  Copyright (c) >NIL: April 2000
 *
 *  NOTES
 *  ~~~~~
 *  "...I've been wanting your undivided attention...
 *   (I like the fact that you're nothing like me)"
 *
 *  Well, this is actually what the VIC20 could say to the
 *  Linux box (I must admit that I was listening to Alanis
 *  Morissette a lot when writing this small program)

 *
 *  DESCRIPTION
 *  ~~~~~~~~~~~
 *  Very fast (about 23.7 KB/s) parallel transfer program
 *  from a Linux PC to a CBM VIC 20, using a simple homemade
 *  parallel cable. See associated docs for details.
 *
 *
 *  CHANGES
 *  ~~~~~~~
 *  v1.1, 2000.04.16
 *    Fixed the buggy timeout processing
 *    Added some transfer rate calculation
 *    Added some command line processing
 *
 *  v1.0, 1999.06.12
 *    Porting of the the Amiga version to the Linux env.
 *
 *
 *  KNOWN BUGS
 *  ~~~~~~~~~~
 *  When the cable is connected to an unpowered VIC, a few bytes
 *  can be send before detecting the error
 *
 */


/* What a bunch of includes for such a small program! */
#include <stdio.h>
#include <unistd.h>
#include <asm/io.h>
#include <sys/io.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>


/* This macro is used for the header transfer */
#define SEND(byte) if (send((byte))) {printf("Cancelling transfer.\n");\
 outb (0x0c, io_base + 2); ioperm (io_base, 3, 0); fclose (fd); exit (1);}

/* Time we'll wait for an _ACK from the VIC in milliseconds */
#define TIMEOUT 200		// 200 milliseconds are long enough

/* Global variables */
FILE *fd;			// The always useful file descriptor
struct timeval tv;		// Let's make those temp global
struct timezone tz;
unsigned short io_base = 0x378;	// Default parallel port base address


/*
 * SEND ROUTINE
 * Sends byte passed as parameter and wait for acknowledgement.
 * Returns 0 on success, -1 on failure.
 */
int
send (unsigned char b)
{
  register int starttime, totaltime;	// For timeout

  /*
   * Write byte to parallel port
   */
  outb (b, io_base);
  /* Activate PC's _STROBE */
  outb (1, io_base + 2);
  /* We don't have to maintain PC's _STROBE low for the
     VIC VIA chip only detects the falling edge...      */
  outb (0, io_base + 2);

  /*
   * Wait for _ACK for TIMEOUT seconds
   */
  if (gettimeofday (&tv, &tz))
    {
      printf ("Error while reading timer.\n");
      return (-1);
    }
  starttime = tv.tv_sec * 1000 + (int) (tv.tv_usec / 1000);

  /* Timeout loop */
  while (inb (io_base + 1) & 0x40)	// Wait for _ACK
    {
      if (gettimeofday (&tv, &tz))
	{
	  printf ("Error while reading timer.\n");
	  return (-1);
	}
      totaltime = tv.tv_sec * 1000 + (int) (tv.tv_usec / 1000) - starttime;
      if (totaltime > TIMEOUT)
	{
	  printf ("Timeout while sending byte.\n");
	  return (-1);
	}
    }
  return (0);			// Success
}


/*
 * MAIN PROGRAM
 */
int
main (int argc, char **argv)
{
  int i;
  int opt_verbose = 0;
  int opt_error = 0;
  unsigned short start, end, size;
  unsigned char byte;
  unsigned int starttime, totaltime;


  /*
   * Command Line Processing
   */
  while ((i = getopt (argc, argv, "vp:")) != -1)
    switch (i)
      {
      case 'v':		// Verbose option
	opt_verbose = -1;
	break;
      case 'p':		// IO Port to be used for parallel transfer
	io_base = strtol (optarg, (char **) NULL, 16);
	if ((io_base != 0x278) && (io_base != 0x378) && (io_base != 0x3bc))
	  {
	    printf
	      ("I/O port address 0x%X is not one of the standard parallel ports.\n",
	       io_base);
	    exit (-1);
	  }
	break;
      case '?':		// Unknown option
	opt_error++;
	break;
      default:
	printf ("?? getopt returned character code 0%o ??\n", i);
      }

  if (((optind + 1) != argc) || opt_error)
    {
      puts ("");
      puts ("cload v1.1 - Linux -> CBM VIC20 program transfer");
      puts ("usage: cload [-v] [-p IO_PORT] file");
      puts ("");
      exit (1);
    }


  /*
   * Initialisation
   */
  if ((fd = fopen (argv[optind], "r")) == NULL)
    {
      printf ("Can't open file %s.\n", argv[optind]);
      exit (1);
    }

  /* Get filesize (there must be a better way to do this) */
  fseek (fd, 0L, SEEK_END);
  /* The first two bytes contain the program start address */
  size = ftell (fd) - 2;
  rewind (fd);

  /* Read first word (=Start Address, in LSB MSB order)       */
  /* WARNING : THIS ONLY APPLIES TO LITTLE ENDIAN MACHINES!!! */
  if (fread (&start, 2, 1, fd) != 1)
    {
      printf ("Error reading start address\n");
      exit (1);
    }
  end = start + size;

  /* Printout some verbose information */
  if (opt_verbose)
    {
      printf ("Program stats:\n");
      printf
	("  start = 0x%04X  (%d)\n  end   = 0x%04X  (%d)\n  size  = 0x%04X  (%d)\n",
	 start, start, end, end, size, size);
    }

  /* Get access to the ports */
  if (ioperm (io_base, 3, 1))
    {
      perror ("ioperm");
      printf ("You have to be root to use this program.\n");
      exit (1);
    }

  // Deactivates bidirectionnality => Output only
  outb (0x00, io_base + 2);

  // Check if the VIC is ready to download
  if (inb (io_base + 1) & 0x40)	// Check _ACK
    {
      printf ("VIC is not ready to download.\n");
      ioperm (io_base, 3, 0);
      fclose (fd);
      exit (1);
    }

  /* INIT is OK, so let's compute some TAGHEUER timings */
  if (gettimeofday (&tv, &tz))
    {
      printf ("Couldn't get time.\n");
      ioperm (io_base, 3, 0);
      fclose (fd);
      exit (1);
    }
  starttime = tv.tv_sec * 1000000 + tv.tv_usec;

  /* SEND HEADER (See VIC 'cload.s' prog for details) */
  for (i = 0; i < 3; i++)
    /* End Of Arrays; Start Of Arrays; Start Of Variables are the same */
    {
      SEND (end >> 8);		// MSB
      SEND (end & 0xff);	// LSB
    }

  /* Start Of Basic */
  SEND (start >> 8);		// MSB
  SEND (start & 0xff);		// LSB

  /*Buffer Pointer (= Start Of Basic - 1) */
  start--;
  SEND (start >> 8);		// MSB
  SEND (start & 0xff);		// LSB

  /* End value for 6502 register Y (One byte only) */
  SEND (size & 0xff);
  /* End Buffer Pointer MSB (One byte only) */
  SEND ((start >> 8) + (size >> 8));

  /*
   * Send program core
   */
  for (i = 0; i < (int) size; i++)
    {
      /* Read one byte and send it */
      if (fread (&byte, 1, 1, fd) != 1)
	{
	  printf ("Could not read from file: Cancelling transfer.\n");
	  break;
	}
      if (send (byte))
	{			// Stop if there is a problem
	  printf ("Cancelling transfer.\n");
	  break;
	}
    }

  /* Even if transfer is not complete, let's see what TAGHEUER says... */
  if (!gettimeofday (&tv, &tz))
    {
      totaltime = tv.tv_sec * 1000000 + tv.tv_usec - starttime;
      printf ("%d bytes transfered in %6.3f ms (%6.3f KB/s)\n", 12 + i,
	      (double) (totaltime / 1000.0),
	      (double) ((12 + i) * 1000000.0 / totaltime / 1024.0));
    }
  else
    printf ("Couldn't get time.\n");

  /*
   * House cleaning
   */
  outb (0x0c, io_base + 2);	// Restore Control register
  ioperm (io_base, 3, 0);
  fclose (fd);
  exit (0);
}
