/*
 *  csave.c  v1.0
 *
 *  Copyright (c) >NIL: April 2000
 *
 *  NOTES
 *  ~~~~~
 *  "...I've been wanting your undivided attention...
 *   (I like the fact that you're nothing like me)"
 *
 *  Well, this is actually what the VIC20 could say to the
 *  Linux box (I must admit that I was listening to Alanis
 *  Morissette a lot when writing this small program)

 *
 *  DESCRIPTION
 *  ~~~~~~~~~~~
 *  Very fast (about 19.5 KB/s) parallel transfer program
 *  from a CBM VIC 20 to a Linux PC, using a simple homemade
 *  parallel cable. See associated docs for details.
 *
 *
 *  CHANGES
 *  ~~~~~~~
 *  v1.0, 2000.04.16
 *    Brand new software
 *
 *
 *  KNOWN BUGS
 *  ~~~~~~~~~~
 *  None yet ;)
 *
 */


/* What a bunch of includes for such a small program! */
#include <stdio.h>
#include <unistd.h>
#include <asm/io.h>
#include <sys/io.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

/* This macro is used for the header transfer */
#define RECEIVE(byte) if (receive((byte))) {printf("Cancelling transfer.\n");\
 outb (0x0c, io_base+2); ioperm (io_base, 3, 0); fclose (fd); exit (1);}

/* Time we'll wait for a _ACK from the VIC in milliseconds */
#define TIMEOUT 200		// 200 milliseconds are long enough

/* Global variables */
FILE *fd;			// The always useful file descriptor
struct timeval tv;		// Let's make those temp global
struct timezone tz;
unsigned short io_base = 0x378;	// Default parallel port base address


/*
 * RECEIVE ROUTINE
 * Reads a byte and copy it at address location given as parameter.
 * Returns 0 on success, -1 on failure.
 */
int
receive (unsigned char *b)
{
  register int starttime, totaltime;	// For timeout

  /*
   * Wait for VIC's _STROBE (= PC's _ACK) for TIMEOUT seconds
   */
  if (gettimeofday (&tv, &tz))
    {
      printf ("Error while reading timer.\n");
      return (-1);
    }
  starttime = tv.tv_sec * 1000 + (int) (tv.tv_usec / 1000);

  /* Timeout loop */
  while (inb (io_base + 1) & 0x40)	// Wait for _ACK
    {
      if (gettimeofday (&tv, &tz))
	{
	  printf ("Error while reading timer.\n");
	  return (-1);
	}
      totaltime = tv.tv_sec * 1000 + (int) (tv.tv_usec / 1000) - starttime;
      if (totaltime > TIMEOUT)
	{
	  printf ("Timeout on receive.\n");
	  return (-1);
	}
    }

  /*
   * Valid data detected => Read byte from parallel port
   */
  *b = inb (io_base);
  /* Acknowledge transfer by activating PC's _STROBE */
  outb (0x21, io_base + 2);	// bit 0x20 = bidirectionnality
  /* We don't have to maintain PC's _STROBE low for the
     VIC VIA chip only detects the falling edge...      */
  outb (0x20, io_base + 2);	// bit 0x20 = bidirectionnality
  return (0);			// Success
}


/*
 * MAIN PROGRAM
 */
int
main (int argc, char **argv)
{
  int i;
  int opt_verbose = 0;
  int opt_error = 0;
  unsigned short start, end, size;
  unsigned char byte;
  unsigned int starttime, totaltime;


  /*
   * Command Line Processing
   */
  while ((i = getopt (argc, argv, "vp:")) != -1)
    switch (i)
      {
      case 'v':		// Verbose option
	opt_verbose = -1;
	break;
      case 'p':		// IO Port to be used for parallel transfer
	io_base = strtol (optarg, (char **) NULL, 16);
	if ((io_base != 0x278) && (io_base != 0x378) && (io_base != 0x3bc))
	  {
	    printf
	      ("I/O port address 0x%X is not one of the standard parallel ports.\n",
	       io_base);
	    exit (-1);
	  }
	break;
      case '?':		// Unknown option
	opt_error++;
	break;
      default:
	printf ("?? getopt returned character code 0%o ??\n", i);
      }

  if (((optind + 1) != argc) || opt_error)
    {
      puts ("");
      puts ("csave v1.0 - CBM VIC20 -> Linux program transfer");
      puts ("usage: csave [-v] [-p IO_PORT] file");
      puts ("");
      exit (1);
    }


  /*
   * Initialisation
   */
  if ((fd = fopen (argv[optind], "w")) == NULL)
    {
      printf ("Can't open file %s.\n", argv[optind]);
      exit (1);
    }

  /* Get access to the ports */
  if (ioperm (io_base, 3, 1))
    {
      perror ("ioperm");
      printf ("You have to be root to use this program.\n");
      exit (1);
    }

  /* Activates bidirectionnality => Input */
  outb (0x20, io_base + 2);

  /* Init is OK, so let's compute some TAGHEUER timings */
  if (gettimeofday (&tv, &tz))
    {
      printf ("Couldn't get time.\n");
      ioperm (io_base, 3, 0);
      fclose (fd);
      exit (1);
    }
  starttime = tv.tv_sec * 1000000 + tv.tv_usec;


  /*
   * Receive Header (See VIC 'csave.s' prog for details)
   */
  /* End Address + 1 (Start Of Variables) in MSB LSB order */
  RECEIVE (&byte);		// I know; you can avoid the using of 'byte'
  end = (int) byte;
  RECEIVE (&byte);
  end = (end << 8) + (int) byte;

  /* Start Address (Start Of Basic) in MSB LSB order */
  RECEIVE (&byte);
  start = (int) byte;
  RECEIVE (&byte);
  start = (start << 8) + (int) byte;
  /* Write Start Address in LSB MSB order                     */
  /* WARNING : THIS ONLY APPLIES TO LITTLE ENDIAN MACHINES!!! */
  if (fwrite (&start, 2, 1, fd) != 1)
  {
    printf ("Error writing start address\n");
    outb (0x0c, io_base + 2);	// Restore Control register
    ioperm (io_base, 3, 0);
    fclose (fd);
    exit (1);
  }
  size = end - start;

  /*
   * Receive program core
   */
  for (i = 0; i < (int) size; i++)
    {
      /* Receive one byte and write it */
      if (receive (&byte))
	{			// Stop if there is a problem
	  printf ("Cancelling transfer.\n");
	  break;
	}
      if (fwrite (&byte, 1, 1, fd) != 1)
	{
	  printf ("Could not write to file: Cancelling transfer.\n");
	  break;
	}
    }

  /* Even if transfer is not complete, let's see what TAGHEUER says... */
  if (!gettimeofday (&tv, &tz))
    {
      totaltime = tv.tv_sec * 1000000 + tv.tv_usec - starttime;
      printf ("%d bytes transfered in %6.3f ms (%6.3f KB/s)\n", 4 + i,
	      (double) (totaltime / 1000.0),
	      (double) ((4 + i) * 1000000.0 / totaltime / 1024.0));
    }
  else
    printf ("Couldn't get time.\n");

  /* Printout some verbose information */
  if (opt_verbose)
    {
      printf ("Program stats:\n");
      printf
	("  start = 0x%04X  (%d)\n  end+1 = 0x%04X  (%d)\n  size  = 0x%04X  (%d)\n",
	 start, start, end, end, size, size);
    }

  /*
   * House cleaning
   */
  outb (0x0c, io_base + 2);	// Restore Control register
  ioperm (io_base, 3, 0);
  fclose (fd);
  exit (0);
}
