; cload.s  v1.1
;
; Very fast (about 23.7 KB/s) parallel transfer program
; from a Linux PC to a CBM VIC 20, using a simple homemade
; parallel cable (See associated doc for cable specs).
;
; This is the 6502 assembly routine to be used with the
; Linux 'cload' program. The corresponding BASIC program is
; given below.
;
; Copyright (c) >NIL: April 2000


	.processor  6502
	.org        267	; Let's use some stack space


; > Initialisation
Init:	lda #$00	; All DDRB Lines set to Input
	sta $9112
	lda #$9e	; Set Pulse Output on VIC _CB2 = PC _ACK on PB Write
	sta $911c
        sta $9110       ; Clean the port and set READY state

; > Receive Header
	ldy #$0c	; Number of Bytes of Header
HdrIn:	jsr InB
	sta $26,y	; Header Organization (Sent in reversed order) :
			; ($27)	- End Buff Ptr MSB     (see below)
			; ($28) - End Y                (see below)
			; ($29) - Buffer Pointer
			;       = SOBasic - 1 at Init, for Y begins at 1
			; ($2B) - Start of Basic       
			; ($2D) - Start of Variables = ($2B) + BuffSize
	dey		; ($2F) - Start of Arrays    = ($2D)
	bne HdrIn	; ($31) - End of Arrays      = ($2D)

; > Receive Loop
	tax		; End Address MSB still in Accumulator -> X
			; Y = 0 at this point (no clearing needed)
Loop:	iny		; Increment Y
	bne Rcv		; No Change of Page needed
	inc $2A		; Change of Page (mod Start Address LSB)
Rcv:	jsr InB		; Get Byte
	sta ($29),y	; Store Indirect at (Buffer Ptr)+Y
	cpy $28		; Compare Y with End Y value
	bne Loop
	cpx $2A		; Compare Current Adress MSB with End Adress MSB (X)
	bne Loop	; Y = Size LSB and BufPtr MSB = EndPtr MSB => Done_it

; > Exit (no cleanup)
	rts


; > Receive a Byte
InB:	lda $911d	; Load IFR
	and #$10	; Test VIC _CB1 bit = PC _STROBE
	beq InB		; Not set => No valid data yet
	lda $9110	; Set => Valid Data loaded into Accumulator
	sta $9110	; Acknowledge byte by sending a pulse on _CB2
	rts


; NOTES
; -----
;
; 1) Pointers calculations are carried out by the host, so that the 
; header's transfer is quite simple. The header, which is sent in 
; reversed order (so that all the pointers are received in Motorola's 
; MSB LSB order) contains the various Start of... ptrs, spared at their 
; predefined locations in VIC's memory, plus 4 more bytes used to store 
; the buffer pointer while receiving, and the data needed to test 
; transfer completion.
;
; 2) The way of writing the data is indirect page zero adressing with
; Y post increment. Though it is slower than self modifing code, I have
; not found a self modifying routine that could be shorter than this
; code, yet.
;
; 3) To test transfer completion, we compare Y and BuffPtr to the ending
; values computed on the host's side.
; Those values are :
; - Real Buffer Size (Start Address Eliminated from file) LSB for Y
; - [Start Address - 1] MSB + BuffSize MSB
;   (and NOT [Start Address - 1 + Buffsize] MSB)
;
; 4) This code has been made the shortest possible, to fit in the stack
; space without disturbing -or being disturbed by- stack data, for we
; are hijacking the end of the stack space.
; Most programs won't go to this space, so you can be fairly sure that
; you'll be able to recall this routine, even after having used a
; program.
;
; 5) This code is only 58 bytes long and I'd really like to hear from
; anyone who can produce a shorter code that does the same thing! ;)
; By the Way, 256-266 corresponds to Floating to ASCII work area, and
; the stack begins at 511, with decreasing pointer.
;
; 6) Unortunaltely, because of the JSRs, this code is not relocatable :(
; You have to run it at adress 267 or modify this assembler file.


; BASIC PROGRAM TO RUN CLOAD ON THE VIC 20
; ----------------------------------------
;
; 10 REM *** CLOAD ***
; 15 REM ** SYS 267 **
; 20 FOR A=267 TO 324
; 30 READ D
; 40 POKE A,D
; 50 NEXT A
; 60 DATA 169,0,141,18,1
; 45,169,158,141,28,145,
; 141,16,145,160,12,32
; 70 DATA 55,1,153,38,0,
; 136,208,247,170,200,20
; 8,2,230,42,32,55
; 80 DATA 1,145,41,196,4
; 0,208,242,228,42,208,2
; 38,96,173,29,145,41
; 90 DATA 16,240,249,173
; ,16,145,141,16,145,96
