; csave.s  v1.0
;
; Very fast (about 19.5 KB/s) parallel transfer program
; from a CBM VIC 20 to a Linux PC, using a simple homemade
; parallel cable (See associated doc for cable specs).
;
; This is the 6502 assembly routine to be used with the
; Linux 'csave' program. The corresponding BASIC program is
; given below.
;
; Copyright (c) >NIL: April 2000


	.processor  6502
	.org        267	; Let's use some stack space


; > Initialisation
Init:	lda #$FF	; All DDRB Lines set to Output
	sta $9112
	lda #$9e	; Set Pulse Output on VIC _CB2 = PC _ACK on PB Write
	sta $911c

; > Send Header
	ldy #$04	; Number of Bytes of Header
HdrOut:	lda $2a,y       ; Start Of Basic & Start Of Variables : $2B to $2E
        sta $26,y       ; Copied into address locations $27 to $2A...
        jsr OutB        ; ...and sent (in reversed order)
        dey
        bne HdrOut

; > Send Loop
Loop:	lda ($27),y     ; Read Indirect at (Buffer Ptr)+Y, with Y=0
        jsr OutB        ; Send byte
	inc $27		; Increment Buffer Ptr LSB
        bne Comp        ; Change of page needed?
        inc $28         ; Increment Buffer Ptr MSB (change of page)
Comp:   lda $29         ; Compare Buffer Ptr with Ending Address + 1
        cmp $27         ; LSB
        bne Loop
        lda $2a
        cmp $28         ; MSB
        bne Loop        ; Buffer Ptr = Ending Address + 1 => Done_it

; > Exit (no cleanup)
	rts

; > Send a Byte
OutB:   sta $9110       ; Sends byte + Handshaking pulse on _CB2
Wait:   lda $911d       ; Load IFR
        and #$10        ; Test VIC _CB1 = PC _STROBE
        beq Wait        ; Not set => byte not acknowledged
        rts             ; Byte acknowledged


; NOTES
; -----
;
; 1) The header, which is sent in reversed order (so that the pointers
; are received in Motorola's MSB LSB order) contain the Ending Address
; +1 (Start Of Variables) and the program Start Address.
; Those addresses are also duplicated in a temporary area, to provide
; the Buffer Pointer which will be used during transfer.
;
; 2) The way of reading the data is indirect page zero adressing with
; Y post increment. In this case, Y is always 0. The Buffer Pointer is
; thus incremented for each byte sent. Though it is slower than self
; modifing code, I have not found a self modifying routine that could
; be shorter than this code, yet.
;
; 3) To test the transfer completion, we simply compare The Buffer
; Pointer to the (copy of) Start Of Variables address.
;
; 4) This code has been made the shortest possible, to fit in the stack
; space without disturbing -or being disturbed by- stack data, for we
; are hijacking the end of the stack space.
; Most programs won't go to this space, so you can be fairly sure that
; you'll be able to recall this routine, even after having used a
; program.
;
; 5) This code is only 59 bytes long and I'd really like to hear from
; anyone who can produce a shorter code that does the same thing! ;)
; By the Way, 256-266 corresponds to Floating to ASCII work area, and
; the stack begins at 511, with decreasing pointer.


; BASIC PROGRAM TO RUN CSAVE ON THE VIC 20
; ----------------------------------------
;
; 10 REM *** CSAVE ***
; 15 REM ** SYS 267 **
; 20 FOR A=267 TO 325
; 30 READ D
; 40 POKE A,D
; 50 NEXT A
; 60 DATA 169,255,141,18
; ,145,169,158,141,28,14
; 5,160,4,185,42,0,153
; 70 DATA 38,0,32,59,1,1
; 36,208,244,177,39,32,5
; 9,1,230,39,208
; 80 DATA 2,230,40,165,4
; 1,197,39,208,239,165,4
; 2,197,40,208,233,96
; 90 DATA 141,16,145,173
; ,29,145,41,16,240,249,
; 96
