/*******************************************************/
/* input64.c                                           */
/* (c) by Andreas Matthies                             */
/* please report bugs and comments to                  */
/* email: andreas.matthies@arcormail.de                */
/*******************************************************/

#include <string.h>
#include <stdio.h>
#include "input64.h"

static struct input64info_t input64info;

static void reset_input64() {
	input64info.middle_count = 0;
	input64info.bits_in_byte = 0;
	input64info.header_count = 0;
	input64info.know_byte_border = 0;
	input64info.actual_byte = 0;
	input64info.actual_bit = 0;
	input64info.next_databyte = 0;
	input64info.parity_count=0;
}


static void init_input64() {
	reset_input64();
}


static void analyse_input64(prg_t *p,tap_t t) {
	char pulsebit[2],pulsebitcount,nextbit,i;
	char bytefull;

	if ((p->state!=headersearch) & (p->type!=prg_input64))
		return;

	if ((1<=t) && (t<=41)) {
		/* short pulse is a Bit0 */
		pulsebitcount = 1;
		pulsebit[0] = 0;
	}
	else if ((42<=t) && (t<=59)) {
		/* middle pulse is Bit1Bit0 or Bit1 */
		pulsebit[0] = pulsebitcount = 1;
		if (++input64info.middle_count % 2) 
			pulsebit[pulsebitcount++] = 0;
	}
	else if (60<=t) {
		/* long pulse is Bit1Bit1 and resets middlecounter */
		pulsebit[0] = pulsebit[1] = 1;
		pulsebitcount = 2;
		input64info.middle_count = 0;
	}
	else {
		/* no input64-pulse */
		pulsebitcount = 0;
		input64info.bits_in_byte = 0;
		reset_input64();
		if ((p->state == headerload) || (p->state == dataload))
			printf("T");
	}

	for (i=0; i<pulsebitcount; i++) {
		nextbit = pulsebit[i];
		/* lsb first */
		input64info.actual_byte >>= 1;
		input64info.actual_byte += nextbit*0x80;
		if (input64info.know_byte_border)
			input64info.actual_bit++;
		bytefull = (input64info.actual_bit == 8);
		if (bytefull)
			input64info.actual_bit = 0;
		if ((p->state==headersearch) || (p->state==datasearch)) {
			if ((p->state==datasearch) && (p->begin>=p->end)) {
				p->state = headersearch;
				p->changestate = 1;
				printf("error- negative size\n");
			}
			if (input64info.actual_byte == 0x16) {
				input64info.header_count++;
				input64info.know_byte_border = 1;
			}
			if (bytefull) {
				/* sync for header */
				if (((input64info.actual_byte & 0xef) == 0x2a)
						&& (p->state == headersearch)
						&& (input64info.header_count >= 11)
					|| (input64info.actual_byte == 0xc5)
						&& (p->state == datasearch)
						&& (input64info.header_count >= 11)) {
					p->type = prg_input64;
					p->databegin = 0;
					p->state++;	/* next state */
					p->changestate = 1;
				}
				else if (input64info.actual_byte != 0x16)
					/* from the beginning */
					reset_input64();
			}
		}
		else {
			/* loading header or data */
			if ((p->state==headerload) && (input64info.next_databyte<25)
				|| (p->state==dataload) && (input64info.next_databyte<(p->end-p->begin)%0x10000))
				input64info.parity_count += nextbit;
			if (bytefull) {
				p->data[input64info.next_databyte++] = input64info.actual_byte;
				if ((p->state==headerload) && (input64info.next_databyte>=27)) {
					/* end of header */
					if ((unsigned char)input64info.parity_count != p->data[25]+p->data[26]*256) {
						/* paraity checksum wrong */
						printf("P");
					}
					/* but we don't care about a wrong char in the filename ;-) */
					p->begin = p->data[18]*256 + p->data[17];
					p->end = p->begin + p->data[20]*256 + p->data[19];
					strncpy((char *)&p->name,(char*)&p->data,16);
					p->state = datasearch;
					p->changestate = 1;
				}
				else if ((p->state==dataload) && (input64info.next_databyte >= (p->end-p->begin)%0x10000+2)) {
					/* end of data-block */
					p->changestate=1;
					if (input64info.parity_count % 0x10000 == p->data[(p->end-p->begin)%0x10000]+p->data[(p->end-p->begin)%0x10000+1]*256)
						p->state = loadok;
					else 
						p->state = loaderror;
				}
			}
		}
	}
}	


struct loader_t input64 = {
	reset_input64,
	init_input64,
	analyse_input64
};
