 
 

#undef	TEST1
#undef	TEST2
#undef	TEST_DELAY

/*********************************************************************/

/* selreg(4) - read low ctrl nibble */

#define ATN_IN          0x10
#define DAV_IN          0x20
#define NRFD_IN         0x08
#define NDAC_IN         0x80

/* selreg(5) - read high ctrl nibble */

#define EOI_IN          0x10
#define IFC_IN          0x20
#define SRQ_IN          0x08
#define REN_IN          0x80

/* selreg(3) - write ctrl byte */
 
#define EOI_OUT         0x01
#define	DAV_OUT		0x02
#define NRFD_OUT        0x04
#define NDAC_OUT        0x08
#define ATN_OUT         0x10
#define DCOUT           0x20    /* high receives ATN/IFC/REN/SRQ, low sends */
#define TEOUT           0x40    /* high sends to bus, low receives */
#define ATNA            0x80

/*********************************************************************/



#define	max(a,b)	(((a)>(b))?(a):(b))

/* Hardware access ----------------------------------------------------*/

static int base = 0;

static int lastreg = -1;
static int read_delay;
static int write_delay;

#include <stdio.h>
#include "iec_hw.h"
#include <asm/io.h>
#include <unistd.h>
#include <errno.h>

#define	DEFAULT_DELAY		20000
#define	PUT_DATA_BYTE(val)	outb((val), base)
#define	READ_HI_NIBBLE()	(inb(base+1) ^ 0x80)
#define	DELAY(del)		while(i--) { READ_HI_NIBBLE(); }

static inline void SELREG(int reg) {
    outb(((reg) ^ 11), base+2);
    lastreg=reg;
}

/* this function avoids read_delay if register already selected */
static inline void SELREG_RDELAY(int reg) {
    if((reg) != lastreg) {
	int i;
	lastreg=(reg);
	outb(((reg) ^ 11), base+2);
	i=read_delay;
	DELAY(i);
    }
}


/* --------------------------------------------------------------------*/

static unsigned char wval[2];


inline void wrb(int reg, int val) {
	int i;

	wval[reg] = val;

/*printf("wrb(reg=%d, val=%02x\n", reg, val);*/

	outb(val, base+0);		/* put value */
	SELREG(reg * 3);		/* select latch */
	i=write_delay; DELAY(i);
	SELREG(4^11);			/* and unselect to save */
}

inline int rdw(int reg)
{
	return(wval[reg]);
}

inline int rdd(void)
{
	int i, tmp;

	SELREG(2);
	i=read_delay; DELAY(i);
	i = READ_HI_NIBBLE();
	tmp = (i & 0xb0) | ((i & 8) ? 0x40 : 0);

	SELREG(1);
	i=read_delay; DELAY(i);
	i = READ_HI_NIBBLE();
	tmp |= ((i >> 4) & 0x0b) | ((i & 8) ? 0x04 : 0);

	SELREG(15);

	return tmp ^ 0xff;
}

inline int rdb_h(void)
{
	/*int i;*/
	SELREG_RDELAY(5);
	/*i=read_delay; DELAY(i);*/
	return READ_HI_NIBBLE();
}

inline int rdb_l(void)
{
	/*int i;*/
	SELREG_RDELAY(4);
	/*i=read_delay; DELAY(i);*/
	return READ_HI_NIBBLE();
}

/*********************************************************************/

static inline void setb(int reg,int bit)
{
	wrb(reg, bit | rdw(reg));
}

static inline void clrb(int reg,int bit)
{
	wrb(reg, (~bit) & rdw(reg));
}

/*********************************************************************/


inline void clrd()
{
	wrb(0,0xff);
}

inline void wrd(int a)
{
	wrb(0,a^0xff);
}

/*********************************************************************/

void settx(void) {
#ifdef TEST1
printf("settx()\n");
#endif
	setb(1,TEOUT);
}

void setrx(void) {
#ifdef TEST1
printf("setrx()\n");
#endif
	clrb(1,TEOUT);
}

void eoihi(void)
{
	setb(1,EOI_OUT);
}

void eoilo(void)
{
	clrb(1,EOI_OUT);
}

void davhi(void)
{
	setb(1,DAV_OUT);
}

void davlo(void)
{
	clrb(1,DAV_OUT);
}

void ndachi(void)
{
	setb(1,NDAC_OUT);
}

void ndaclo(void)
{
	clrb(1,NDAC_OUT);
}

void nrfdhi(void)
{
	setb(1,NRFD_OUT);
}

void nrfdlo(void)
{
	clrb(1,NRFD_OUT);
}

void atnahi(void)
{
	clrb(1,ATNA);
}

void atnalo(void)
{
	setb(1,ATNA);
}

void ledaus(void)
{
}

void ledan(void)
{
}

/*********************************************************************/

inline int davishi()
{
	return (rdb_l() & DAV_IN);
}

inline int eoiishi()
{
	return (rdb_h() & EOI_IN);
}

inline int nrfdishi()
{
	return (rdb_l() & NRFD_IN);
}

inline int ndacishi()
{
	return (rdb_l() & NDAC_IN);
}

inline int atnishi()
{
#ifdef TEST1
	int _atnishi(void);
	int val = _atnishi();
	if(!val) {
	  printf("ATN is low (val=%02x)\n",val);
	} 
	return val;
}

int _atnishi()
{
#endif
	return (rdb_l() & ATN_IN);
}

/*********************************************************************/

int davislo(void) 
{
	return !davishi();
}

int ndacislo(void) 
{
	return !ndacishi();
}

int nrfdislo(void) 
{
	return !nrfdishi();
}

int eoiislo(void) 
{
	return !eoiishi();
}

int atnislo(void) 
{
	return !atnishi();
}

/*********************************************************************/

#define	BUFLEN	10000

int measure_read(int val) {
	int j, i;
	int sample[BUFLEN];
	int want;

	/* measure read delay:
	 *
	 * 1) set output data register to val, with long delays
	 * 2) unselect latch
	 * 3) select data nibble to read and wait until value is 
	 *    ok
	 */

	PUT_DATA_BYTE(0x7d);		/* DAV active -> latch enable 
					   but also TE active -> sending 
					   -> data loop from out to in */
	SELREG(3);
	i=DEFAULT_DELAY; DELAY(i);
	SELREG(15);			/* and unselect to save */

	/* write data register */

#ifdef TEST_DELAY
	printf("Get read delay: write $%02x\n",val);
#endif
	PUT_DATA_BYTE(val);		/* put value */
	SELREG(0);			/* select latch */
	i=DEFAULT_DELAY; DELAY(i);
	SELREG(15);			/* and unselect to save */

	i=DEFAULT_DELAY; DELAY(i);

	want = ((val<<4) & 0xb0) | ((val & 4) ? 8 : 0);

#ifdef TEST_DELAY
	printf("pre-read: %02x, want %02x\n", READ_HI_NIBBLE(), want);
#endif
	i=0;
	SELREG(1);			/* select low nibble */
	do {
	    sample[i++] = READ_HI_NIBBLE();
	} while(i<BUFLEN);

	for(j=0, i=0;i<BUFLEN;i++) {
	    if( (sample[i] ^ want) & 0xb8 ) j=i; 
	}
	if(j>=BUFLEN-1) {
	    return -1;
	}
#ifdef TEST_DELAY
	printf("needed %d tries, first values were %02x, %02x, %02x\n",
		j, sample[0], sample[1], sample[2]);
#endif
	
	return j;
}

int measure_write(int val, int rdel) {
	int j, i, tmp, want;
	int sample[BUFLEN];

	/* measure write delay:
	 *
	 * 1) set output data register to ~val, with long delays
	 * 2) unselect latch
	 * 3) set output data register to val, with certain delay
	 * 4) after read_delay check if data is taken
	 * 5) if data is not ok, increase write delay and go back to 1)
	 */


	PUT_DATA_BYTE(0x7d);
	SELREG(3);
	i=DEFAULT_DELAY; DELAY(i);
	SELREG(15);			/* and unselect to save */

#ifdef TEST_DELAY
	printf("Get write delay: write $%02x\n",val);
#endif
	want = ((val<<4) & 0xb0) | ((val & 4) ? 8 : 0);

	j = 0;
	do {

	    /* write data register */
	    PUT_DATA_BYTE(val^0xff);	/* put value */
	    SELREG(0);			/* select latch */
	    i=DEFAULT_DELAY; DELAY(i);
	    SELREG(15);			/* and unselect to save */
	    i=DEFAULT_DELAY; DELAY(i);

	    PUT_DATA_BYTE(val);		/* put value */
	    SELREG(0);
	    i=j; DELAY(i);
	    SELREG(15);

	    SELREG(1);
	    i=rdel; DELAY(i);

	    tmp = READ_HI_NIBBLE();

	    /* printf("%d: tmp%02x, want=%02x\n",j,tmp,want); */

	    if( (tmp ^ want) & 0xb8) {
		j++;			/* different -> increase delay */
	    } else {
		break;			/* ok -> end loop */
	    }
	} while(j<DEFAULT_DELAY);
#ifdef TEST_DELAY
	printf("get after %d tries\n",j);
#endif

	return j;
}

int measure_delay(int *rdel, int *wdel) {
	int del1, del2, del3, del4;

	del1 = measure_read(0x55);
	del2 = measure_read(0xaa);
	del3 = measure_read(0x00);
	del4 = measure_read(0xff);

	if (del1 < 0 || del2 < 0 || del3 < 0 || del4 < 0) {
	    fprintf(stderr, "lptiec: Did not find hardware...!!!\n");
	    return -1;
	}

	*rdel = max( del1, del2 );
	*rdel = max( *rdel, del3);
	*rdel = max( *rdel, del4);

	*rdel = *rdel * 1.5 + 10;

	del1 = measure_write(0x55, *rdel);
	del2 = measure_write(0xaa, *rdel);
	del3 = measure_write(0x00, *rdel);
	del4 = measure_write(0xff, *rdel);

	*wdel = max( del1, del2 );
	*wdel = max( *wdel, del3);
	*wdel = max( *wdel, del4);

	*wdel = *wdel * 1.5 + 10;

	return 0;
}

void iec_setup(void) {
#ifdef TEST2
	printf("setup: ctrl->$3f, \n");
#endif
	wrb(1, 0x3f);	/* control port */
	wrb(0, 0xff);	/* data port */

}

int iecini(int baseaddr) {
	int er;
	uid_t uid, euid;

	uid = getuid();
	euid= geteuid();
	
	base = baseaddr;

	if(ioperm(base, 3, -1)<0) {
	    printf("lptiec: could not get io permissions to port %x\n", base);
	    printf("ioperm: errno=%d (%s)\n",errno,strerror(errno));
	    exit(1);
	}

	if(chroot(".")<0) {
	    printf("lptiec: Could not chroot to base directory!\n");
	    printf("chroot: errno=%d (%s)\n",errno,strerror(errno));
	    exit(1);
	}

	printf("Resetting permissions to normal user!\n");
	if(setuid(uid) < 0) {
	    printf("lptiec: could not set priviliges to user %d\n", uid);
	    printf("setuid: errno=%d (%s)\n",errno,strerror(errno));
	    exit(1);
	}
	
	er = measure_delay(&read_delay, &write_delay);

	if(er<0) {
	    return -1;
	}

	printf("set read delay to %d\n", read_delay);
	printf("set write delay to %d\n", write_delay);

	iec_setup();

	return 0;
}



