/*******************************************************/
/* novaload.c                                          */
/* (c) by Andreas Matthies                             */
/* please report bugs and comments to                  */
/* email: andreas.matthies@arcormail.de                */
/*******************************************************/

#include <string.h>
#include <stdio.h>
#include "novaload.h"

static struct novaloadinfo_t novaloadinfo;


static void reset_novaload() {
	novaloadinfo.bit0_in_a_row = 0;
	novaloadinfo.checksum = 0;
	novaloadinfo.byteborder = 0;
	novaloadinfo.actualbyte = 0;
	novaloadinfo.actualbit = 0;
	novaloadinfo.byteborder = 0;
	novaloadinfo.nextdatabyte = 0;
	novaloadinfo.subblockcounter = 0;
}


static void init_novaload() {
	reset_novaload();
}


static void analyse_novaload(prg_t *p,tap_t t) {
	unsigned char thisbit,bytefull;
	unsigned short i; 
	/* Bit 0:  511 T-States; Tap-Values: 21-50
	   Bit 1: 1222 T-States; Tap-Values: 71-100 */

	if ((p->state!=headersearch) && (p->type!=prg_novaload))
		return;

	if ((21<=t) && (t<=50)) {
		/* short pulse is a Bit0 */
		thisbit = 0;
	}
	else if ((71<=t) && (t<=100)) {
		/* long pulse is a Bit1 */
		thisbit = 1;
	}
	else {
		/* no novaload-pulse */
		thisbit = 2;
		if ((p->state == headerload) || (p->state == dataload)) {
			novaloadinfo.actualbit=0;
		}
	}
	if (thisbit<2) {
		novaloadinfo.actualbyte >>=1;
		novaloadinfo.actualbyte += thisbit*0x80;
		novaloadinfo.actualbyte &= 0xff;
		if (novaloadinfo.byteborder)
			novaloadinfo.actualbit++;
	}
	bytefull = (novaloadinfo.actualbit==8);
	if (bytefull)
		novaloadinfo.actualbit = 0;
	
	if (p->state==datasearch) {
		/* Data follows immediatly after Header */
		p->state = dataload;
		p->changestate = 1;
		p->databegin = 0;
		novaloadinfo.nextdatabyte = 0;
	}
	
	if (p->state==headersearch) {
		/* searching novaload-file */
		if ((thisbit==1) && (novaloadinfo.bit0_in_a_row>254)) {
			/* Novaload-trailer found */
			novaloadinfo.byteborder=1;
		}
		if (thisbit==0)
			novaloadinfo.bit0_in_a_row++;
		else
			novaloadinfo.bit0_in_a_row = 0;
		if (bytefull) {
			if (novaloadinfo.actualbyte==0xaa) {
				/* found flag */
				p->state = headerload;
				p->changestate = 1;
				novaloadinfo.actualbit = 0;
				p->type = prg_novaload;
			}
			else {
				/* wrong flag */
				reset_novaload();
			}
		}
	}
	else {
		/* loading nova-file */
		if (thisbit>1) {
			printf("T");
			p->state = loaderror;
			p->changestate = 1;
		}
		if (bytefull) {
			p->data[novaloadinfo.nextdatabyte++] = novaloadinfo.actualbyte;
			if (p->state==headerload) {
				unsigned char filelen = p->data[0];
				if (filelen==0x55) {
					novaloadinfo.novatype = 1; /* special Nova-Block */
					printf("NOVA-special block, not implemented! Sorry, ");
					/* for -e: create a dummy-file */
					p->begin = p->end = 0;
					strncpy(p->name,"NOVASPECIAL     ",16);
					p->changestate = 1;
					p->state = loaderror;
				}
				else if ((unsigned char)novaloadinfo.nextdatabyte==filelen+7) {
					/* header loaded */
					p->begin = p->data[filelen+1] + p->data[filelen+2]*256 + 0x100;
					p->end = p->data[filelen+3] + p->data[filelen+4]*256;
					strncpy((char *)&p->name,(char*)&p->data[1],filelen);
					for (i=filelen; i<16; i++)
						p->name[i] = 0x20;
					p->state = datasearch;
					novaloadinfo.novatype = 0; /* normal Nova-Block */
					p->changestate = 1;
				}
			}
			if ((p->state==dataload) && (novaloadinfo.novatype==0)) {
				/* loading normal nova-file */
				/* test for end */
				if (novaloadinfo.nextdatabyte >= (unsigned)((p->end-p->begin)%0x10000+1)) {
					if (novaloadinfo.actualbyte==novaloadinfo.checksum)
						p->state = loadok;
					else
						p->state = loaderror;
					p->changestate = 1;
				}
				if (!novaloadinfo.subblockcounter) {
					/* checksumbyte */
					novaloadinfo.nextdatabyte--;
					if (novaloadinfo.checksum!=p->data[novaloadinfo.nextdatabyte])
						printf("C");
				}
				novaloadinfo.subblockcounter++;
				if (novaloadinfo.subblockcounter>256)
					novaloadinfo.subblockcounter = 0;
			}
			novaloadinfo.checksum += novaloadinfo.actualbyte;
		}
	}
}


struct loader_t novaload = {
	reset_novaload,
	init_novaload,
	analyse_novaload
};
