/*******************************************************/
/* tape64.h                                            */
/* (c) by Andreas Matthies                             */
/* please report bugs and comments to                  */
/* email: andreas.matthies@arcormail.de                */
/*******************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "loader.h"

#define VERSION "V1.2"
#define LAST_CHANGE "07/09/1999"

#define MAXINPUTBUFFER 1000000
#define MAXTAPBUFFER 100000
#define MAXSAMPLE 256
#define MAXTAPVALUE 256*256*256-1

typedef unsigned short sample_t;
typedef enum {up,down} direction_t;

typedef char BOOL;

typedef struct {
	enum {sample, tap, unknown} inputtype;
	long sample_freq;
	unsigned long sample_width;
	unsigned char sample_channels;
	unsigned char tap_version;
	sample_t trigamp;
	char inputname[255], tapname[255], p00dirname[255], t64name[255];
	char name_noextension[255];
	BOOL automatic_yes;
	BOOL p00output;
	BOOL t64output;
	BOOL tapoutput;
	BOOL p00numerate;
	BOOL reverse_wave;
	BOOL save_loaderrors;
	unsigned long tap_v0_delay_ms;
	double tapespeed;
} param_t;


const char TAP_IDENTIFIER[12] = "C64-TAPE-RAW";
typedef struct {
	char identifier[12];
	unsigned char version;
	unsigned char future[3];
	unsigned long filesize;
} tapheader_t;

const char T64_IDENTIFIER[32] = "C64S tape image file\0\0\0\0\0\0\0\0\0\0\0\0";
typedef struct {
	char identifier[32];
	char major_version, minor_version;
	short max_entries, used_entries;
	short notused;
	char tape_name[24];
} t64header_t;

typedef struct {
	unsigned char c64stype;
	unsigned char filetype;
	unsigned short begin;
	unsigned short end;
	unsigned short notused;
	unsigned long offset;
	unsigned long evennotused;
	unsigned char name[16];
} t64dir_t;


typedef struct cbm {
	unsigned short begin;
	unsigned short end;
	unsigned char c64name[16];
	unsigned char data[MAXPRGSIZE];
	struct cbm *next;
} cbmlist_t;


const char allowedfilechars[] = " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!$%&()=',.-_;~+";
const double c64_pulse = 985248.0;

typedef struct {
	sample_t last_sample, sample;
	sample_t last_min, last_max;
	direction_t last_direction, direction;
	unsigned long sample_count, last_trigger;
} samplestate_t;
