#ifdef WIN32
# include <windows.h>
# include <mmsystem.h>
#endif /* WIN32 */
#include <string.h>
#include <stdio.h>
#include "tapir.h"
#include "types.h"
#include "sound.h"

#define BUFFERSIZE 8*1024
#define NUM_BUF 8

struct threaddata {
  HANDLE availbuf;
  HANDLE texit;
};

struct threaddata tdata;

HWND        hWnd;	/* Handle to main window */
HINSTANCE  hInst;	/* Handle to instance data*/

HWAVEIN indev;
HMIXER mixdev;

WAVEHDR wavebuf[NUM_BUF];
int curbuf;
char *allbuf;

HANDLE availbuf;
HANDLE texit;

HANDLE recthread;
DWORD recthreadid;

sound_info info;

void process_buffer(void *data, int length)
{
    int i;
    short *sdata = (short *) data;
    double *ddata;
    int slength = length / 2;
    int temp = 0;
    double thresh = 400;
    DWORD wrbytes;
    static int state = 0;

    static double pulselen = 0;
    static double lastdata = 0;

    ddata = LocalAlloc(0, slength*sizeof(double));
    if (!ddata)
        return;

    filter_buffer(sdata, ddata, slength, 1);
//    convert_buffer(sdata, ddata, slength, 1);

    for (i=0; i<slength; i++)
    {
        if (state == 0) /* positive pulse */
        {
            if (ddata[i] <= -thresh)
                state = 1;
        }
        else /* negative pulse */
        {
            if (ddata[i] >= thresh)
            {
                double deltalen;

                /* Do a linear interpolation between the last and
                   current points */
                deltalen = (thresh - lastdata) / (ddata[i] - lastdata);
                pulselen += deltalen;

                write_tap_data(pulselen, 44100);

                state = 0;

                /* This is the time between the threshold crossing and
                   the current sample */
                pulselen = -deltalen;
            }
        }
        pulselen+=1;
        lastdata=ddata[i];
    }

    LocalFree(ddata);
}

#ifdef OLD_SCHOOL
DWORD WINAPI recfunc (LPVOID data)
{
    struct threaddata *ptdata = (struct threaddata *) data;
    HANDLE availbuf = ptdata->availbuf;
    HANDLE texit = ptdata->texit;
    int prevbuf;
    HANDLE waitvec[2]={availbuf, texit};

    while (1)
    {
        if (WaitForSingleObject(availbuf, INFINITE) == WAIT_OBJECT_0)
        {
            /* Check if we're asked to exit */
            if (WaitForSingleObject (texit, 0) == WAIT_OBJECT_0)
                break;
  
            /* Poll the next buffer */
            while (wavebuf[curbuf].dwFlags & WHDR_DONE)
            {
                waveInUnprepareHeader(indev, &wavebuf[curbuf],
                                      sizeof(WAVEHDR));
                
                process_buffer(wavebuf[curbuf].lpData,
                               wavebuf[curbuf].dwBufferLength);
                
                prevbuf=curbuf-1;
                if (prevbuf < 0)
                    prevbuf = NUM_BUF-1;
                
                curbuf++;
                if (curbuf >= NUM_BUF)
                    curbuf = 0;
                
                wavebuf[prevbuf].lpData = &allbuf[prevbuf*BUFFERSIZE];
                wavebuf[prevbuf].dwBufferLength = BUFFERSIZE;
                wavebuf[prevbuf].dwFlags = 0;
                wavebuf[prevbuf].dwLoops = 0;
                waveInPrepareHeader(indev, &wavebuf[prevbuf], sizeof(WAVEHDR));
                waveInAddBuffer(indev, &wavebuf[prevbuf], sizeof(WAVEHDR));
            }
            ResetEvent(availbuf);
        }
    }
    return 0;
}
#endif /* OLD_SCHOOL */

/* Callback for sound input */
int sound_process_input(WAVEHDR *header)
{
	process_buffer(header->lpData, header->dwBytesRecorded);
	return TRUE;
}

/* Callback for sound output. Not used yet. */
int sound_process_output(WAVEHDR *header)
{
	return TRUE;
}

int start_recording (void)
{
#ifdef OLD_SCHOOL

    MMRESULT res;

    res = waveInStart(indev);
//    printf ("waveInStart == %d\n", res);

#else

	sound_start_input(&info);

#endif /* OLD_SCHOOL */
}

int stop_recording (void)
{
#ifdef OLD_SCHOOL

    waveInStop(indev);

#else

	sound_stop_input(&info);

#endif /* OLD_SCHOOL */
}

void perror_wavein(MMRESULT res)
{
    char ollon[300];
    waveInGetErrorText(res, ollon, 300);
    puts(ollon);
}

void perror_message(uint32 error)
{
    char ollon[300];

    FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM,
                   NULL, error, 0, ollon, 300, NULL);
    puts(ollon);
}

int init_stuff (void)
{
    int i;

#ifdef OLD_SCHOOL /* Old way of doing sound in */

	MMRESULT res;
    WAVEFORMATEX wfx;

    /* Event that will be signalled when there are samples available */
    availbuf = CreateEvent(NULL, TRUE, FALSE, "Buffer available");
    if (!availbuf)
    {
        perror_message (GetLastError());
        return 0;
    }

    /* Event that will be signalled when the thread is to be exited */
    texit = CreateEvent(NULL, TRUE, FALSE, "Exit thread");
    if (!texit)
    {
        perror_message (GetLastError());
        return 0;
    }

    wfx.wFormatTag = WAVE_FORMAT_PCM;
    wfx.nChannels = 1;
    wfx.nSamplesPerSec = 44100;
    wfx.wBitsPerSample = 16;
    wfx.nAvgBytesPerSec = 44100 * 1 * 16 / 8;
    wfx.nBlockAlign = 1 * 16 / 8;
    wfx.cbSize = 0;

    res = waveInOpen (&indev, WAVE_MAPPER, &wfx,
                      (DWORD) availbuf, 0, CALLBACK_EVENT);
    if (res != MMSYSERR_NOERROR)
    {
        perror_wavein(res);
        return 0;
    }
    
/*
    res = mixerOpen (&mixdev, indev, 0, 0, MIXER_OBJECTF_HWAVEIN);
    if (res != MMSYSERR_NOERROR)
    {
        perror_wavein(res);
        return 0;
    }
*/

    /* Allocate and prepare waveform buffers */   
    allbuf = LocalAlloc (LMEM_FIXED, NUM_BUF*BUFFERSIZE);
    if (!allbuf)
    {
        perror_message (GetLastError());
        return 0;
    }

    for (i = 0; i < NUM_BUF-1; i++)
    {
        wavebuf[i].lpData = &allbuf[i*BUFFERSIZE];
        wavebuf[i].dwBufferLength = BUFFERSIZE;
        wavebuf[i].dwFlags = 0;
        wavebuf[i].dwLoops = 0;
        
        res = waveInPrepareHeader (indev, &wavebuf[i], sizeof(WAVEHDR));
        if (res != MMSYSERR_NOERROR)
        {
            perror_wavein(res);
            return 0;
        }
//        printf ("waveInAddBuffer %p->%p\n", &wavebuf[i], wavebuf[i].lpData);
        res = waveInAddBuffer (indev, &wavebuf[i], sizeof(WAVEHDR));
        if (res != MMSYSERR_NOERROR)
        {
            perror_wavein(res);
            return 0;
        }
     
        curbuf = 0;

    }

    /* Tell receiving thread about the events */
    tdata.availbuf = availbuf;
    tdata.texit = texit;

    /* Create the thread that will receive the audio */
    recthread = CreateThread (NULL, 0, recfunc, &tdata, CREATE_SUSPENDED,
                              &recthreadid);
    if (!recthread)
    {
        perror_message (GetLastError());
        return 0;
    }

    if (!SetThreadPriority (recthread, THREAD_PRIORITY_ABOVE_NORMAL))
    {
        perror_message (GetLastError());
        // return 0;
    }

    if (ResumeThread(recthread) == 0xFFFFFFFF)
    {
        perror_message (GetLastError());
        return 0;
    }

#else /* Newer way */

	info.samplerate = 44100;
	info.channels = MONO;
	info.sample_format = SIXTEEN_BITS;
	info.in = NULL;
	info.out = NULL;
	info.format = NULL;
	info.in_handle = 0;
	info.in_callback = sound_process_input;
	info.out_callback = sound_process_output;
	for(i = 0; i < NUM_BUFFERS; i++ ) info.header[i] = NULL;
	
	sound_choose_input(-1, "Aureal", &info);
	/* sound_choose_output(-1, "Aureal", &info); */

	sound_set_format(&info);
	sound_open_input(&info);

#endif /* OLD_SCHOOL */
}


int close_stuff (void)
{
#ifdef OLD_SCHOOL

    MMRESULT res;
    int i;
    
    /* Currently, recording must be running in order for the thread
       to notice the texit flag */
    start_recording();

    /* Tell thread to exit */
    SetEvent (texit);
    /* Wait for thread to exit */
    while (WaitForSingleObject (recthread, 1000) == WAIT_TIMEOUT)
    {
        /* Do nothing */
    }

    waveInReset(indev);
    res = waveInClose(indev);
    
    CloseHandle (recthread);
    CloseHandle (texit);
    CloseHandle (availbuf);

    for (i = 0; i < NUM_BUF; i++)
        waveInUnprepareHeader (indev, &wavebuf[i], sizeof(WAVEHDR));
    LocalFree(allbuf);

#else

	sound_stop_input(&info);
	sound_close_input(&info);

#endif /* OLD_SCHOOL */
}

void print_current_time(void)
{
    char ollon[100];
    SYSTEMTIME systime;

    GetLocalTime (&systime);
    GetTimeFormat (LOCALE_USER_DEFAULT, 0, &systime, NULL, ollon, 100);
    puts (ollon);
}

int main (void)
{
    MSG msg;
    int i;

    if (!open_tap_file("testfile.tap"))
        return (FALSE);

    if (!init_stuff())
        return (FALSE);
    
    print_current_time();

    start_recording();
    
    /* Wait for return */
    {
        HANDLE hstdin=GetStdHandle(STD_INPUT_HANDLE);
        char buf[100];
        DWORD nread;
        ReadFile(hstdin, buf, 1, &nread, NULL);
    }
        
    close_stuff();
    
    print_current_time();

    if (!close_tap_file())
        return (FALSE);
    
    return 0;
}

