/*
 * This will someday be a nice library for sound I/O on different platforms
 * Right now only the input works (and only for win32)
 */

#ifdef WIN32
# include <windows.h>
# include <process.h>
#elif defined LINUX
# include <sys/types.h>
# include <sys/stat.h>
# include <sys/ioctl.h>
# include <sys/soundcard.h>
# include <fcntl.h>
# include <errno.h>
# include <pthread.h>
# include <unistd.h>
#endif /* WIN32 */

#include <malloc.h>
#include <stdio.h>
#include <string.h>

#include "tapir.h"
#include "sound.h"
#include "types.h"

#ifdef WIN32
DWORD WINAPI sound_input_thread(sound_input *in)
{
    int prevbuf;

    DBG("sound_input_thread();\n");
    while (1) {
        if (WaitForSingleObject(in->in_event, 1000) == WAIT_OBJECT_0) {
			
            DBG("in_event signaled\n");
            /* Check if we're asked to exit */
            if (WaitForSingleObject (in->in_exit, 0) != WAIT_OBJECT_0) {

		/* Poll the next buffer */
                while (in->header[in->current_buffer]->dwFlags & WHDR_DONE) {
                    waveInUnprepareHeader(in->handle, 
					  in->header[in->current_buffer],
					  sizeof(WAVEHDR));
					
                    if (in->playing_flag == TRUE) {
                        DBG("Calling input callback\n");
                        in->callback(in->header[in->current_buffer]);
                        DBG("Callback finished\n");
                    }

                    prevbuf = in->current_buffer - 1;

                    if (prevbuf < 0)
                        prevbuf = NUM_BUFFERS - 1;
					
                    in->current_buffer++;
                    if (in->current_buffer >= NUM_BUFFERS)
                        in->current_buffer = 0;
					
                    in->header[prevbuf]->lpData =
			in->buffer +
			BUFFER_SIZE * in->format->nBlockAlign * prevbuf;

                    in->header[prevbuf]->dwBufferLength =
			BUFFER_SIZE * in->format->nBlockAlign;
					
                    in->header[prevbuf]->dwFlags = 0;
                    in->header[prevbuf]->dwLoops = 0;
					
                    waveInPrepareHeader(in->handle,
					in->header[prevbuf],
					sizeof(WAVEHDR));
                    waveInAddBuffer(in->handle,
				    in->header[prevbuf],
				    sizeof(WAVEHDR));
                }
            }
            ResetEvent(in->in_event);
        }
        if(WaitForSingleObject (in->in_exit, 0) == WAIT_OBJECT_0)
            break;
        DBG("wait timed out\n");
    }
	
    DBG("ending sound_input_thread();\n");
    return 0;
}

#elif defined LINUX
void* sound_input_thread(void *arg)
{
    uint32 len;
/*    audio_errinfo errinfo; */
    sound_input *in = arg;
    
    DBG("sound_input_thread();\n");
    sleep(1);

    if((in->buffer = malloc(in->frag_size)) == NULL) {
	perror(__FUNCTION__);
	in->exit_flag = TRUE;
    }

    while (in->exit_flag == FALSE) {
	/* Read from the input */
	if((len = read(in->fd, in->buffer, in->frag_size)) == -1) {
	    perror(__FUNCTION__);
	    in->exit_flag = TRUE;
	} else if(in->playing_flag == TRUE) {
/*          Only available in OSS 3.9.4 and later
            ioctl(in->fd, SNDCTL_DSP_GETERROR, &errinfo);
	    if(errinfo.rec_overruns > 0) DBG("Recording overrun detected\n");
*/
	    DBG("Calling input callback();\n");
	    in->callback(in, len);
	    DBG("callback finished\n");
	}
    }
    free(in->buffer);

    DBG("ending sound_input_thread();\n");
    return 0;
}
#endif

int sound_open_input(sound_input *in) 
{
#ifdef WIN32
    MMRESULT res;
    int i;

    DBG("sound_open_input();\n");

    if(in == NULL) return FALSE;

    if(NULL == (in->in_event = CreateEvent(NULL, TRUE, FALSE, NULL))) {
        DBG("error with CreateEvent (in_event)\n");
	return FALSE;
    }
    if(NULL == (in->in_exit = CreateEvent(NULL, TRUE, FALSE, NULL))) {
        DBG("error with CreateEvent (in_exit)\n");
        return FALSE;
    }
    in->playing_flag = FALSE;
/*	if(MMSYSERR_NOERROR != (res = waveInOpen( &info->in_handle,
	info->in_ID,
	info->format,
	(DWORD)info->in_event,
	0,
	CALLBACK_EVENT |
	WAVE_FORMAT_DIRECT))) {

	It seems WAVE_FORMAT_DIRECT is unpopular with win95/98.
	Thanks Micro$oft for not bothering to put this in the
	documentation...
*/
    if(MMSYSERR_NOERROR != (res = waveInOpen( &in->handle,
					      in->in_ID,
					      in->format,
					      (DWORD)in->in_event,
					      0,
					      CALLBACK_EVENT ))) {
	DBG("Can't open input!\n");
#ifdef _DEBUG
	/* Some extra debug info used to find out why this wouldn't
	   work on win95 */
	switch(res) {
	case MMSYSERR_ALLOCATED :
	    DBG("MMSYSERR_ALLOCATED\n");
	    break;
	case MMSYSERR_BADDEVICEID :
	    DBG("MMSYSERR_BADDEVICEID\n");
	    break;
	case MMSYSERR_NODRIVER :
	    DBG("MMSYSERR_NODRIVER\n");
	    break;
	case MMSYSERR_NOMEM :
	    DBG("MMSYSERR_NOMEM\n");
	    break;
	case WAVERR_BADFORMAT :
	    DBG("WAVERR_BADFORMAT\n");
	    break;
	case MMSYSERR_INVALFLAG :
	    DBG("WAVERR_INVALFLAG\n");
	    break;
	default :
	    DBG("Unknown error!\n");
	}
#endif
        return FALSE;
    }
	
    if(NULL != (in->buffer = LocalAlloc(LMEM_FIXED,
					   NUM_BUFFERS * BUFFER_SIZE *
					   in->format->nBlockAlign))) {
        for(i = 0; i < NUM_BUFFERS; i++) {
            if(NULL != (in->header[i] =
			LocalAlloc(LMEM_FIXED, sizeof(WAVEHDR)))) {
		
                in->header[i]->lpData = in->buffer +
		    BUFFER_SIZE * in->format->nBlockAlign * i;
                in->header[i]->dwBufferLength =
		    BUFFER_SIZE * in->format->nBlockAlign;
                in->header[i]->dwUser = i;
                in->header[i]->dwFlags = 0;
                in->header[i]->dwLoops = 0;
                if (i < NUM_BUFFERS -1) {
                    if(MMSYSERR_NOERROR !=
		       (res = waveInPrepareHeader(in->handle,
						  in->header[i], 
						  sizeof(WAVEHDR)))) {
                        DBG("Can't prepare wave header!\n");
                        goto sound_open_input_error;
                    }
                    if(MMSYSERR_NOERROR !=
		       (res = waveInAddBuffer(in->handle,
					      in->header[i],
					      sizeof(WAVEHDR)))) {
                        DBG("Can't add buffer to input!\n");
                        goto sound_open_input_error;
                    }
                }
            }
        }
        in->current_buffer = 0;
		
        if(-1 == (int)(in->in_thread_handle =
		       CreateThread(NULL,
				    0,
				    (void*)sound_input_thread,
				    in,
				    CREATE_SUSPENDED,
				    &(in->in_thread_id)))) {
            DBG("CreateThread(); error!\n");
            return FALSE;
        }
        if (!SetThreadPriority (in->in_thread_handle,
				THREAD_PRIORITY_ABOVE_NORMAL))
        {
            DBG("SetThreadPriority(); error!\n");
            return FALSE;
        }
        if (ResumeThread(in->in_thread_handle) == 0xFFFFFFFF)
        {
            DBG("Error with ResumeThread();\n");
            CloseHandle(in->in_thread_handle);
            goto sound_open_input_error;
        }
        Sleep(100);
		
        return TRUE;
		
        /* And now, a label to GOTO. Nice, Eh? :) */
sound_open_input_error:
        /* Something went wrong */
        if(in->header[0]->lpData != NULL) {
            LocalFree(in->header[0]->lpData);
        }
        for(i = 0; i < NUM_BUFFERS; i++) {
            if(in->header[i] != NULL) {
                LocalFree(in->header[i]);
                in->header[i] = NULL;
                in->header[i]->lpData = NULL;
            }
        }
    }

    return FALSE;

#elif defined LINUX

    int temp_format;
    int temp_channels;
    int temp_rate;

    DBG("sound_open_input();\n");

    if(in == NULL || in->name == NULL) return FALSE;

    in->playing_flag = FALSE;

    if((in->fd = open(in->name, O_RDONLY, 0)) == -1) {
	perror(in->name);
	return FALSE;
    }

    /* Set sample format */
    temp_format = in->format;
    if(ioctl(in->fd, SNDCTL_DSP_SETFMT, &(in->format)) == -1) {
	perror(__FUNCTION__);
	return FALSE;
    }
    if(in->format != temp_format) {
	fprintf(stderr, "%s doesn't seem to support that format.\n", in->name);
	return FALSE;
    }

    /* Set number of channels */
    temp_channels = in->channels;
    if(ioctl(in->fd, SNDCTL_DSP_CHANNELS, &(in->channels)) == -1) {
	perror(__FUNCTION__);
	return FALSE;
    }
    if(in->channels != temp_channels) {
	fprintf(stderr, "%s doesn't seem to support %d channels.\n",
		in->name, temp_channels);
	return FALSE;
    }

    /* Set sample rate */
    temp_rate = in->samplerate;
    if(ioctl(in->fd, SNDCTL_DSP_SPEED, &in->samplerate) == -1) {
	perror(__FUNCTION__);
	return FALSE;
    }
    /* A few percent difference is nothing to care about here */
    if((((float)in->samplerate / temp_rate) < 0.98) ||
       (((float)in->samplerate / temp_rate) > 1.02)) {
	fprintf(stderr, "%s doesn't seem to support %dHz sample rate.\n",
		in->name, temp_rate);
	return FALSE;
    }

    if(ioctl(in->fd, SNDCTL_DSP_GETBLKSIZE, &in->frag_size) == -1) {
	in->frag_size = 0;
	perror(__FUNCTION__);
    }

    /* Create the input thread */
    in->exit_flag = FALSE;
    if(pthread_create(&(in->thread), NULL, sound_input_thread, in) != 0) {
	perror(__FUNCTION__);
	in->exit_flag = TRUE;
	return FALSE;
    }

    return TRUE;
#endif
}

int sound_close_input(sound_input *in)
{
#ifdef WIN32
    MMRESULT res;
    int i;
	
    DBG("sound_close_input();\n");
	
    if(in == NULL) return FALSE;

    SetEvent(in->in_exit);
    while (WaitForSingleObject(in->in_thread_handle, 1500) == WAIT_TIMEOUT) {
        /* Do nothing */
        DBG("Timed out waiting for input thread to finish\n");
    }
	
    DBG("waveInReset();\n");
    waveInReset(in->handle);
	
    if(MMSYSERR_NOERROR != (res = waveInClose( in->handle ))) {
	DBG("Can't close input: ");
	if(res == MMSYSERR_INVALHANDLE) {
	    DBG("invalid handle\n");
	} else {
	    DBG("unknown error\n");
	}
    }

    CloseHandle(in->in_event);
    CloseHandle(in->in_exit);
    CloseHandle(in->in_thread_handle);
	
    if(in->header[0]->lpData != NULL) {
        LocalFree(in->header[0]->lpData);
    }
    for(i = 0; i < NUM_BUFFERS; i++) {
        if(in->header[i] != NULL) {
            waveInUnprepareHeader (in->handle,
				   in->header[i],
				   sizeof(WAVEHDR));
            in->header[i]->lpData = NULL;
            LocalFree(in->header[i]);
            in->header[i] = NULL;
        }
    }
#elif defined LINUX

    DBG("sound_close_input();\n");

    if(in == NULL) return FALSE;

    /* Tell the sound input threead to stop and wait for it to do so */
    if(in->exit_flag == FALSE) {
	in->exit_flag = TRUE;
	pthread_join(in->thread, NULL);
    }
    
    if(close(in->fd) == -1) {
	perror(__FUNCTION__);
	return FALSE;
    }
#endif
    return TRUE;
}

int sound_start_input(sound_input* in)
{
    DBG("sound_start_input();\n");
#ifdef WIN32
    if(MMSYSERR_NOERROR != waveInStart(in->handle)) {
        DBG("waveInStart(); error\n");
        return FALSE;
    }
#elif defined LINUX

#endif
    in->playing_flag = TRUE;
    return TRUE;
}

int sound_stop_input(sound_input* in)
{
    DBG("sound_stop_input();\n");
    in->playing_flag = FALSE;
#ifdef WIN32
    if(MMSYSERR_NOERROR != waveInStop(in->handle)) {
        DBG("waveInStop(); error\n");
        in->playing_flag = TRUE;
        return FALSE;
    }
#elif defined LINUX

#endif
    return TRUE;
}

int sound_set_format(sound_input* in, sound_output* out)
{
    DBG("sound_set_format();\n");

    if(in == NULL) return FALSE;

#ifdef WIN32
    if(in->format == NULL) {
	if(NULL == (in->format = malloc(sizeof(WAVEFORMATEX)))) {
	    return FALSE;
	}
    }

    in->format->wFormatTag = WAVE_FORMAT_PCM;
    in->format->nChannels = (in->channels == MONO) ? 1 : 2;
    in->format->wBitsPerSample =
	(in->sample_format == EIGHT_BITS) ? 8 : 16;
    in->format->nBlockAlign =
	in->format->nChannels * ((int)in->format->wBitsPerSample / 8);
    in->format->nSamplesPerSec = in->samplerate;
    in->format->nAvgBytesPerSec =
	in->format->nSamplesPerSec * in->format->nBlockAlign;
    in->format->cbSize = 0;
#elif defined LINUX
    in->format = (in->sample_format == EIGHT_BITS) ?
	AFMT_U8 : AFMT_S16_NE;
#endif
    return TRUE;
}

int sound_choose_input(int32 input_num, char *input_name, sound_input* in)
{
#ifdef WIN32
    UINT num_inputs;
    MMRESULT res;
    UINT i = 0;
#elif defined LINUX
    int namelen;
#endif
    DBG("sound_choose_input();\n");

    if(in == NULL) return FALSE;

#ifdef WIN32
    if(0 == (num_inputs = waveInGetNumDevs())) {
        DBG("No wave input devices present!\n");
	return FALSE;
    }

    DBGPRINTF("number of devices: %i\n",num_inputs);

    if(NULL == (in->caps = malloc(sizeof(WAVEINCAPS)))) {
	return FALSE;
    }

    if(input_num == -1 && input_name == NULL) {
        if(MMSYSERR_NOERROR ==
	   (res = waveInGetDevCaps(WAVE_MAPPER, in->caps,
				   sizeof(WAVEINCAPS)))) {
	    DBGPRINTF("input #%i: ", i);
	    DBGPRINTF("%s\n", in->caps->szPname);
	    /* OK, we've found the right input */
	    in->in_ID = -1;
	    return TRUE;
        }
    }
    for(i = 0; i < num_inputs; i++) {
        if(MMSYSERR_NOERROR ==
	   (res = waveInGetDevCaps(i, in->caps, sizeof(WAVEINCAPS)))) {
	    DBGPRINTF("input #%i: ", i);
	    DBGPRINTF("%s\n", in->caps->szPname);
	    /* Match attributes of this input */
	    if(((input_num != -1) &&
		((int32)i == input_num)	)
	       ||
	       ((input_name != NULL) &&
		(0 == strncmp(input_name,
			      in->caps->szPname,
			      strlen(input_name))))) {
				
				/* OK, we've found the right input */
		in->in_ID = i;
		return TRUE;
	    }
        }
    }

    free(in->caps);
    in->caps = NULL;

#elif defined LINUX

    if(input_num == -1 && input_name == NULL) {
	namelen = strlen("/dev/dsp") + 1;
	if(NULL == (in->name = malloc(namelen))) {
	    fprintf(stderr, "Can't allocate memory in sound_choose_input();\n");
	    return FALSE;
	}
	strncpy(in->name, "/dev/dsp", namelen);
    } else if(input_name != NULL) {
	/* Use the already supplied name */
	namelen = strlen(input_name) + 1;
	if(NULL == (in->name = malloc(namelen))) {
	    fprintf(stderr, "Can't allocate memory in sound_choose_input();\n");
	    return FALSE;
	}
	strncpy(in->name, input_name, namelen);
    } else if(input_num != -1) {
	/* OK, so this is a bit ugly... */
	char input_num_txt[0x10] = "";

	snprintf(input_num_txt, 0x10, "%d", input_num);
	namelen = strlen("/dev/dsp") + strlen(input_num_txt) + 1;
	if(NULL == (in->name = malloc(namelen))) {
	    fprintf(stderr, "Can't allocate memory in sound_choose_input();\n");
	    return FALSE;
	}
	strcpy(in->name, "/dev/dsp");
	strcat(in->name, input_num_txt);
    }

    DBGPRINTF("input name: %s\n", in->name);

    return TRUE;

#endif

    return FALSE;
}

int sound_choose_output(int32 output_num, char *output_name, sound_output* out)
{
#ifdef WIN32
    UINT num_outputs;
    MMRESULT res;
    UINT i;
#elif defined LINUX

#endif

    DBG("sound_choose_output();\n");

    if(out == NULL) return FALSE;

#ifdef WIN32
    if(0 == (num_outputs = waveOutGetNumDevs())) {
        fprintf(stderr, "No wave output devices present!\n");
        return FALSE;
    }
	
    DBGPRINTF("number of devices: %i\n",num_outputs);

    if(NULL == (out->caps = malloc(sizeof(WAVEOUTCAPS)))) {
	return FALSE;
    }
	
    for(i = 0; i < num_outputs; i++) {
        if(MMSYSERR_NOERROR ==
	   (res = waveOutGetDevCaps(i, out->caps, sizeof(WAVEOUTCAPS)))) {
            DBGPRINTF("output #%i: ", i);
	    DBGPRINTF("%s\n", out->caps->szPname);
	    /* Match attributes of this output */
	    if(((output_num != -1) && ((int32)i == output_num)) ||
	       ((output_name != NULL) &&
		(0 == strncmp(output_name,
			      out->caps->szPname,
			      strlen(output_name))))) {
				/* OK, we've found the right output */
		out->out_ID = i;
		printf("Output chosen: %s\n", out->caps->szPname);
		return TRUE;
	    }
        }
    }

    free(out->caps);
    out->caps = NULL;

#elif defined LINUX

    return TRUE;

#endif

    return FALSE;
}
