
#ifndef _SOUND_H		  
#define _SOUND_H

#ifdef WIN32
# include <windows.h>
# include <mmsystem.h>
#elif defined LINUX
# include <sys/soundcard.h>
# include <sys/ioctl.h>
# include <errno.h>
# include <fcntl.h>
# include <pthread.h>
#else
# error Sorry, target system is not supported.
#endif

#include "types.h"

/* Just print debug info when compiled for debugging */
/* Rather crappy, but useful... */
#ifdef _DEBUG
char crap_debug_string[0x100];
# include <stdio.h>
# ifdef WIN32
#  define DBG(x) OutputDebugString(x);
#  define DBGPRINTF(x,y) _snprintf(crap_debug_string, 0xff, (x), (y)); OutputDebugString(crap_debug_string);
#  define DBGERROR FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, \
                   GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), \
                   crap_debug_string, 0xff, NULL); \
	           OutputDebugString(crap_debug_string);
# elif defined LINUX
#  define DBG(x) fprintf(stderr, (x));
#  define DBGPRINTF(x,y) snprintf(crap_debug_string, 0xff, (x), (y)); fputs(crap_debug_string, stderr);
#  define DBGERROR perror(__FUNCTION__);
# endif
#else
# define DBG(x)
# define DBGPRINTF(x,y)
# define DBGERROR
#endif

/* Sample format defines */
#define MONO 1
#define STEREO 2

/* These aren't used (yet) */
#define MONO_LEFT 3
#define MONO_RIGHT 4
#define MONO_AUTO 5
#define MONO_BOTH 6

#define EIGHT_BITS 0
#define SIXTEEN_BITS 1

#ifdef WIN32
/* Well, 32k-samples should be enough, right? */
# define BUFFER_SIZE (32*1024)
# define NUM_BUFFERS 4
/*#elif defined LINUX
  # define BUFFER_SIZE (32*1024) */
#endif

typedef struct {
    uint32 samplerate;
    uint8 sample_format;
    char *buffer;
    int channels;
    int	playing_flag;

#ifdef WIN32
    int	(*callback)(LPWAVEHDR);
    int	current_buffer;
    UINT		in_ID;
    WAVEINCAPS		*caps;
    HWAVEIN		handle;
    HANDLE		in_event;
    HANDLE		in_exit;
    HANDLE		in_thread_handle;
    DWORD		in_thread_id;
    LPWAVEFORMATEX	format;
    LPWAVEHDR		header[NUM_BUFFERS];
#elif defined LINUX
    int	(*callback)(void*, uint32);
    int fd;
    char* name;
    int format;
    int frag_size;
    pthread_t thread;
    int	exit_flag;
#endif /*WIN32 */
} sound_input;

typedef struct {
    uint32      samplerate;
    int		channels;
    uint8	sample_format;
    char	*buffer;
    int		current_buffer;
    int		playing_flag;

#ifdef WIN32
    int	(*callback)(LPWAVEHDR);
    UINT		out_ID;
    WAVEOUTCAPS		*caps;
    HWAVEOUT		handle;
    HANDLE		out_event;
    HANDLE		out_exit;
    HANDLE		out_thread_handle;
    DWORD		out_thread_id;
    LPWAVEFORMATEX	format;
    LPWAVEHDR		header[NUM_BUFFERS];
#elif defined LINUX
    int	(*callback)(void*, uint32);
    int fd;
    char* name;
    int format;
    int frag_size;
    pthread_t thread;
    int	exit_flag;
#endif /* WIN32 */
} sound_output;

int sound_open_input(sound_input* in);
int sound_close_input(sound_input* in);
int sound_start_input(sound_input* in);
int sound_stop_input(sound_input* in);
int sound_set_format(sound_input* in, sound_output* out);
int sound_choose_input(int32 input_num,
		       char *input_name,
		       sound_input* in);
int sound_choose_output(int32 output_num,
			char *output_name,
			sound_output* out);
#endif /* _SOUND_H */
