#ifdef WIN32
# include <windows.h>
#else
# include <errno.h>
#endif /* WIN32 */
#include <stdio.h>
#include "tapir.h"
#include "types.h"

static FILE* tapfile;

/*static unsigned int tapbytes; */
static uint32 tapbytes;

void write_tap_data (double pulselen, uint32 fs)
{
    /* One sample corresponds to 985248/samplerate cycles (PAL) */
    uint32 cycles = (uint32)((double)pulselen * settings.clock_freq / fs);

    if (cycles <= 255*8)
    {
        uint8 tapdata = (uint8)((cycles+4)/8);
        fwrite(&tapdata, 1, 1, tapfile);
        tapbytes++;
    } else {
        uint8 tapdata[4];
        tapdata[0]=0;
        tapdata[1]=(uint8)(cycles & 0xFF);
        tapdata[2]=(uint8)((cycles>>8) & 0xFF);
        tapdata[3]=(uint8)((cycles>>16) & 0xFF);
        fwrite(&tapdata, 4, 1, tapfile);
        tapbytes+=4;
    }
}

int open_tap_file(char *name)
{
    char signature[]="C64-TAPE-RAW";
    uint8 reserved[4]={1, 0, 0, 0};

    tapfile = fopen(name, "wb");

    if(tapfile == NULL) {
        PERROR;
	return FALSE;
    }

    if(fwrite(&signature, 1, 12, tapfile) != 12
       || fwrite(&reserved, 1, 4, tapfile) != 4) {

        PERROR;
	fclose(tapfile);
	return FALSE;
    }

    if(fwrite(&reserved, 1, 4, tapfile) != 4) {
        PERROR;
	fclose(tapfile);
	return FALSE;
    }

    return TRUE;
}

int close_tap_file(void)
{
    uint8 length[4]={0, 0, 0, 0};

    length[0] = (uint8)(tapbytes&0xff);
    length[1] = (uint8)((tapbytes>>8)&0xff);
    length[2] = (uint8)((tapbytes>>16)&0xff);
    length[3] = (uint8)((tapbytes>>24)&0xff);
    
    if(fseek(tapfile, 0x10, SEEK_SET) != 0) {
        PERROR;
	return FALSE;
    } else if(fwrite(&length, 1, 4, tapfile) != 4) {
        PERROR;
	return FALSE;
    } else if(fclose(tapfile) == EOF) {
        PERROR;
	return FALSE;
    }
    return TRUE;
}




