#ifndef _TAPIR_H
#define _TAPIR_H

#ifdef WIN32
# include <windows.h>
#endif /* WIN32 */

#include "types.h"
#include "cmdline.h"

/* Constants (from mtap) */
#define C64 0
#define VIC20 1
#define C16 2

#define PAL 0
#define NTSC 1

#define TAP 0
#define TURBO 1
#define KERNAL 2

#define C64PAL 985248
#define C64NTSC 1022727
#define VIC20PAL 1108405
#define VIC20NTSC 1022727
#define C16PAL 886724
#define C16NTSC 894886

/* A struct for keeping global settings.
   Feels better to have them all in one place. */
struct tapir_settings {
  uint8 type;
  uint8 video;
  uint8 loader;
  uint32 clock_freq;
} settings;

extern struct gengetopt_args_info options;

/* tapir.c */
#ifdef WIN32
void perror_message (uint32 error);
#elif defined LINUX
void perror_message (int error);
#endif

/* output.c */
void write_output_data (double pulselen, uint32 fs);
int open_output_file (char *name);
int close_output_file (void);

/* tap.c */
void write_tap_data (double pulselen, uint32 fs);
int open_tap_file (char *name);
int close_tap_file (void);

/* turbotape.c */
void write_turbo_data (double pulselen, uint32 fs);
int open_turbo_file (char *name);
int close_turbo_file (void);

/* filter.c */
void filter_buffer (short *indata, double *outdata, int length, int invert, uint8 sample_format);
void convert_buffer (short *indata, double *outdata, int length, int invert);

#endif /* _TAPIR_H */
