#ifdef WIN32
# include <windows.h>
#endif /* WIN32 */

#include "tapir.h"
#include "types.h"
#include <ctype.h>
#include <stdio.h>
#include <string.h>

#define THRESHOLD 263
#define TOOSHORT 180
#define TOOLONG 380

/* This should be expanded quite a bit in the future, maybe even #ifdef:ed
   so that we can have different strings for win32 and linux */
#define FORBIDDEN_CHARS "*?<>:|/\\\""

#define GOTO_STATE(state) {decode_state = (state);}
/* printf ("Going to state %s.\n", statename[(state)]);} */

static FILE* turbofile;

enum states {SYNC, SYNC9, SEQUENCE, BLOCKTYPE, HEADER, 
             HEADER2, HEADER3, HEADER4, HEADER5, FILENAME, 
             FILENAME_CHECKSUM, DATA, DATA_CHECKSUM};

/* Used for debugging purposes */
char *statename[]={"SYNC", "SYNC9", "SEQUENCE", "BLOCKTYPE", "HEADER",
                   "HEADER2", "HEADER3", "HEADER4", "HEADER5", "FILENAME",
                   "FILENAME_CHECKSUM", "DATA", "DATA_CHECKSUM"};

static enum states decode_state = SYNC;

/* A (not a very good) filename translation function. Right now it only
   transforms all non-printable characters (and those in FORBIDDEN_CHARS)
   to underscore */
char* safe_filename(char* filename)
{
    char* tempptr = filename;
    char tempstr[2] = " ";

    do {
	tempstr[0] = *tempptr;
	if(!isprint(*tempptr) || (strpbrk(tempstr, FORBIDDEN_CHARS) != NULL)) {
	    *tempptr = '_';
	}
    } while(*(++tempptr) != '\0');

    return filename;
}

void write_turbo_data(double pulselen, uint32 fs)
{
    static unsigned char byte;	/* The byte that is being decoded */
    static int bitcnt;		/* How many bits have been read */
    static int seqbyte;		/* Which byte the countdown expects */
    static char filename_buffer[200];
    static char* filename;
    static unsigned char checksum;
    static int datalength;
    static uint32 start_address;
    static uint32 end_address;

    uint32 cycles = (uint32) ((double)pulselen * settings.clock_freq / fs);
    int bit;

    if (cycles < TOOSHORT || cycles > TOOLONG)
    {
        close_turbo_file();
        GOTO_STATE(SYNC);
        return;
    }

    bit = cycles < THRESHOLD ? 0 : 1;

    byte = byte << 1 | bit;
    bitcnt++;

    if (decode_state == SYNC)
    {
        if (byte == 0x02)
        {
            bitcnt = 0;
            GOTO_STATE(SYNC9);
        }
    }
    else
    {
        if (bitcnt == 8)
        {
            bitcnt = 0;
            switch (decode_state)
            {
            case SYNC9:
                if (byte == 0x09)
                {
                    seqbyte = 8;
                    GOTO_STATE(SEQUENCE);
                }
                else if (byte != 0x02)
                    GOTO_STATE(SYNC);
                break;
            case SEQUENCE:
                if (byte == seqbyte)
                {
                    if (--seqbyte == 0)
                        GOTO_STATE(BLOCKTYPE);
                }
                else
                    GOTO_STATE(SYNC);
                break;
            case BLOCKTYPE:
                if (byte == 0)
                {
                    checksum = 0;
                    datalength = (int) (end_address - start_address);
                    if (turbofile)
                    {
                        fputc(start_address & 0xFF, turbofile);
                        fputc(start_address >> 8, turbofile);
                    }
                    printf ("Loading.\n");
                    GOTO_STATE(DATA);
                }
                else if (byte == 1 || byte == 2)
                {
                    GOTO_STATE(HEADER);
                }
                else
                    GOTO_STATE(SYNC);
                break;
            case HEADER:
                start_address = byte;
                GOTO_STATE(HEADER2);
                break;
            case HEADER2:
                start_address |= byte*256;
                GOTO_STATE(HEADER3);
                break;
            case HEADER3:
                end_address = byte;
                GOTO_STATE(HEADER4);
                break;
            case HEADER4:
                end_address |= byte*256;
                GOTO_STATE(HEADER5);
                break;
            case HEADER5:
                /* byte not used */
                datalength = 32;  /* does not follow Hkan Sundell's description */
                checksum = 0;
                filename = filename_buffer;
                GOTO_STATE(FILENAME);
                break;
            case FILENAME:
                *filename++ = byte;
                checksum ^= byte;
                if (--datalength == 0)
                {
                    /* skip trailing spaces in filename */
                    while (*--filename == 0x20)
                        ;
                    *++filename = '\0';
                    
		    /* Transform the filename to a safe one */
                    safe_filename(filename_buffer);

                    printf ("Found %s ($%04X-$%04X).\n",
                            filename_buffer, start_address, end_address);
                    strcat(filename_buffer, ".prg");
                    open_turbo_file(filename_buffer);                        
                    GOTO_STATE(FILENAME_CHECKSUM);
                }
                break;
            case FILENAME_CHECKSUM:
                GOTO_STATE(SYNC);
                break;
            case DATA:
                if (turbofile)
                    fputc(byte, turbofile);
                checksum ^= byte;
                if (--datalength == 0)
                {
                    GOTO_STATE(DATA_CHECKSUM);
                }
                break;
            case DATA_CHECKSUM:
                close_turbo_file();
                if (byte != checksum)
                {
                    printf("Checksum error in data.\n");
                }
                printf("Ready.\n");
                GOTO_STATE(SYNC);
                break;
            default:
                printf ("Unknown state.\n");
                GOTO_STATE(SYNC);
                break;
            }
        }
    }
}

int open_turbo_file(char *name)
{
    turbofile = fopen(name, "wb");

    if(turbofile == NULL) {
        PERROR;
        return FALSE;
    }

    decode_state = SYNC;

    return TRUE;
}

int close_turbo_file(void)
{
    if (turbofile)
    {
        if(fclose(turbofile) == EOF) {
	    PERROR;
            return FALSE;
        }
        turbofile = NULL;
    }
    return TRUE;
}
