/*******************************************************/
/* turbotape.c                                         */
/* (c) by Andreas Matthies                             */
/* please report bugs and comments to                  */
/* email: andreas.matthies@arcormail.de                */
/*******************************************************/

#include <string.h>
#include <stdio.h>
#include "turbotape.h"

static struct ttinfo_t ttinfo;

/* msb first */
static const unsigned char ttbit[8] = 
	{0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};

/* Header */
static const unsigned char ttheaderbyte[9] =
	{9,8,7,6,5,4,3,2,1};
	
static const unsigned char ttprgnamebegin = 6;
static const unsigned char ttprgdatabegin = 1;
static unsigned char ttheader[72];


static void reset_turbotape() {
	ttinfo.nextheaderbyte = ttinfo.lastheaderbyte = 0;
	ttinfo.nextbit = 0;
	ttinfo.actualbyte = 0;
	ttinfo.datachksum = 0;
}



static void init_turbotape() {
	int byte, bit;
	reset_turbotape();
	for (byte=0;byte<9;byte++)
		for (bit=0;bit<8;bit++)
			ttheader[byte*8+bit] = (ttheaderbyte[byte]&ttbit[bit]?1:0);
}


static void analyse_turbotape(prg_t *p,tap_t t) {
	
	char actualbit; 
	char bytefull;
	char couldbeheader;
	long i;

	if ((p->state!=headersearch)&&(p->type!=prg_turbotape))
		return;

	/* analyse trigger */
	if ((18<=t) && (t<=33))
		/* Bit 0 */
		actualbit = 0;
	else if ((34<=t) && (t<=49))
		/* Bit 1 */
		actualbit = 1;
	else 
		/* no bit of a turbotape coding */
		actualbit = 2;

	/* append bit to actualbyte (only for loading states) */
	bytefull = 0;
	if (actualbit>1) {
		ttinfo.nextbit = 0;
		ttinfo.actualbyte = 0;
	}
	else if ((actualbit==1) || (actualbit==0)) {
		if (ttinfo.nextbit<=7) {
			ttinfo.actualbyte = ((actualbit==1)?
				ttinfo.actualbyte | ttbit[ttinfo.nextbit]:
				ttinfo.actualbyte & (255-ttbit[ttinfo.nextbit]));
			ttinfo.nextbit++;
			if (ttinfo.nextbit==8) {
				bytefull = 1;
				ttinfo.nextbit = 0;
			}
		}
	}
	if ((p->state==datasearch) && (p->begin>=p->end)) {
		p->state = headersearch;
		p->changestate = 1;
	}
	/* searching for header */
	if ((p->state==headersearch)||(p->state==datasearch)) {
		if (actualbit>1) {
			ttinfo.nextbit = 0;
			ttinfo.nextheaderbyte = ttinfo.lastheaderbyte = 0;
		}
		else {
			ttinfo.header[ttinfo.nextheaderbyte] = actualbit;
			ttinfo.nextheaderbyte = (ttinfo.nextheaderbyte+1) % 72;
			couldbeheader = (ttinfo.nextheaderbyte==ttinfo.lastheaderbyte);
			if (couldbeheader) {
				ttinfo.lastheaderbyte = (ttinfo.lastheaderbyte+1) % 72;
				i = 0;
			}
			/* 72 consequtive bits collected -> check for header */
			while (couldbeheader && (i<72)) {
				couldbeheader = couldbeheader &&
					(ttheader[i]==ttinfo.header[(ttinfo.nextheaderbyte+i)%72]);
					i++;
			}
			if (couldbeheader) {
				/* now it's sure: is is a header!!! */
				p->type = prg_turbotape;
				p->databegin = ttprgdatabegin;
				p->state++;	/* next state */
				p->changestate = 1;
				ttinfo.nextbit = 0; /* reset bitcounter */
				ttinfo.nextdatabyte = 0;
			}
		}
	}
	else {
		/* loading header/data */
		if (actualbit>1) {
			printf("T");
			p->changestate = 1;
			p->state = loaderror;
		}
		if (bytefull) {
			/* byte is full */
			p->data[ttinfo.nextdatabyte++] = ttinfo.actualbyte;
			if (p->state==headerload) {
				if (ttinfo.nextdatabyte && !p->data[0])
					/* sorry, no tt-header */
					p->state = headersearch;
				/* check for end of header */
				if (ttinfo.nextdatabyte>=ttprgnamebegin+16) {
					/* end of header */
					p->begin = p->data[2]*256 + p->data[1];
					p->end = p->data[4]*256 + p->data[3];
					strncpy((char *)&p->name,(char*)&p->data[ttprgnamebegin],16);
					p->state = datasearch;
					p->changestate = 1;
				}
			}
			else if ((p->state==dataload) && bytefull) {
				/* check for end of data-block */
				if (ttinfo.nextdatabyte > (p->end-p->begin)%0x10000+ttprgdatabegin) {
					p->changestate=1;
					ttinfo.chksum = p->data[ttinfo.nextdatabyte-1];
					if (ttinfo.datachksum==ttinfo.chksum)
						p->state = loadok;
					else
						p->state = loaderror;

				}
				else if (ttinfo.nextdatabyte>ttprgdatabegin) {
					/* checksum */
					ttinfo.datachksum ^= p->data[ttinfo.nextdatabyte-1];
				}
			}
		}
	}
}

struct loader_t turbotape = {
	reset_turbotape,
	init_turbotape,
	analyse_turbotape
};
